import click

from sqlite3 import connect
from os import remove, path
from sys import exit
from shutil import copyfile

from pandas import concat, read_sql
from numpy import power

from .preprocess import uniprot, net, sec, quantification, normalization, meta, query
from .score import monomer, scoring
from .learn import pyprophet, combine
from .quantify import quantitative_matrix, enrichment_test
from .plot import plot_features, check_sqlite_table
from .export import export_tables

from pyprophet.data_handling import transform_threads, transform_pi0_lambda

# import cProfile

@click.group(chain=True)
@click.version_option()
def cli():
    """
    SECAT: Size-Exclusion Chromatography Algorithmic Toolkit. test

    Visit https://github.com/grosenberger/secat for usage instructions and help.
    """

# SECAT import data
@cli.command()
@click.argument('infiles', nargs=-1, type=click.Path(exists=True))
@click.option('--out', 'outfile', required=True, type=click.Path(exists=False), help='Output SECAT file.')
# Reference files
@click.option('--sec', 'secfile', required=True, type=click.Path(exists=True), help='The input SEC calibration file.')
@click.option('--net', 'netfile', required=False, type=click.Path(exists=True), help='Reference binary protein-protein interaction file in STRING-DB or HUPO-PSI MITAB (2.5-2.7) format.')
@click.option('--posnet', 'posnetfile', required=False, type=click.Path(exists=True), help='Reference binary positive protein-protein interaction file in STRING-DB or HUPO-PSI MITAB (2.5-2.7) format.')
@click.option('--negnet', 'negnetfile', required=False, type=click.Path(exists=True), help='Reference binary negative protein-protein interaction file in STRING-DB or HUPO-PSI MITAB (2.5-2.7) format.')
@click.option('--uniprot', 'uniprotfile', required=True, type=click.Path(exists=True), help='Reference molecular weights file in UniProt XML format.')
@click.option('--columns', default=["run_id","sec_id","sec_mw","condition_id","replicate_id","run_id","protein_id","peptide_id","peptide_intensity"], show_default=True, type=(str,str,str,str,str,str,str,str,str), help='Column names for SEC & peptide quantification files')


# Parameters for normalization
@click.option('--normalize/--no-normalize', default=True, show_default=True, help='Normalize quantification data by sliding window cycling LOWESS normaklization.')
@click.option('--normalize_window','normalize_window', default=5, show_default=True, type=int, help='Number of SEC fractions per sliding window.')
@click.option('--normalize_padded/--no-normalize_padded', default=True, show_default=True, help='Use padding for first and last SEC fractions.')
# Parameters for decoys
@click.option('--decoy_intensity_bins', 'decoy_intensity_bins', default=1, show_default=True, type=int, help='Number of decoy bins for intensity.')
@click.option('--decoy_left_sec_bins', 'decoy_left_sec_bins', default=1, show_default=True, type=int, help='Number of decoy bins for left SEC fraction.')
@click.option('--decoy_right_sec_bins', 'decoy_right_sec_bins', default=1, show_default=True, type=int, help='Number of decoy bins for right SEC fraction.')
@click.option('--decoy_oversample','decoy_oversample', default=2, show_default=True, type=int, help='Number of iterations to sample decoys.')
@click.option('--decoy_subsample/--no-decoy_subsample', default=False, show_default=True, help='Whether decoys should be subsampled to be approximately of similar number as targets.')
# @click.option('--decoy_exclude/--no-decoy_exclude', default=True, show_default=True, help='Whether decoy interactions also covered by targets should be excluded.')
@click.option('--min_interaction_confidence', 'min_interaction_confidence', default=0.0, show_default=True, type=float, help='Minimum interaction confidence for prior information from network.')
@click.option('--interaction_confidence_bins', 'interaction_confidence_bins', default=100, show_default=True, type=int, help='Number of interaction confidence bins for grouped error rate estimation.')
@click.option('--interaction_confidence_quantile/--no-interaction_confidence_quantile', default=True, show_default=True, help='Whether interaction confidence bins should be grouped by quantiles.')
@click.option('--use_cached_uniprot', 'cache', default=True, required=False, show_default=True, type=bool, help='Whether to use the Uniprot table parsed from a previous run')
def preprocess(infiles, outfile, secfile, netfile, posnetfile, negnetfile, uniprotfile, columns, normalize, normalize_window, normalize_padded, decoy_intensity_bins, decoy_left_sec_bins, decoy_right_sec_bins, decoy_oversample, decoy_subsample, min_interaction_confidence, interaction_confidence_bins, interaction_confidence_quantile, cache): # decoy_exclude
    """
    Import and preprocess SEC data.
    """

    # Prepare output file
    try:
        remove(outfile)
    except OSError:
        pass

    con = connect(outfile)

    # Generate SEC definition table
    click.echo("Info: Parsing SEC definition file %s." % secfile)
    sec_data = sec(secfile, columns)
    sec_data.to_df().to_sql('SEC', con, index=False)

    # Generate Peptide quantification table
    run_ids = sec_data.to_df()['run_id'].unique() # Extract valid run_ids from SEC definition table

    quantification_list = []
    for infile in infiles:
        click.echo("Info: Parsing peptide quantification file %s." % infile)
        quantification_list.append(quantification(infile, columns, run_ids).to_df())
    quantification_data = concat(quantification_list)

    # Normalize quantitative data
    if normalize:
        click.echo("Info: Normalizing quantitative data.")
        quantification_data = normalization(quantification_data, sec_data.to_df(), normalize_window, normalize_padded, outfile).to_df()

    # Store quantification data
    quantification_data.to_sql('QUANTIFICATION' ,con, index=False, if_exists='append')

    # Generate peptide and protein meta data over all conditions and replicates
    click.echo("Info: Generating peptide and protein meta data.")
    meta_data = meta(quantification_data, sec_data.to_df(), decoy_intensity_bins, decoy_left_sec_bins, decoy_right_sec_bins)
    meta_data.peptide_meta.to_sql('PEPTIDE_META', con, index=False)
    meta_data.protein_meta.to_sql('PROTEIN_META', con, index=False)

    # Generate UniProt table
    click.echo("Info: Parsing UniProt XML file %s." % uniprotfile)
    uniprot_data = uniprot(uniprotfile, cache)
    uniprot_data.to_df().to_sql('PROTEIN', con, index=False)

    # Generate Network table
    if netfile != None:
        click.echo("Info: Parsing network file %s." % netfile)
    else:
        click.echo("Info: No reference network file was provided.")
        # decoy_exclude = False
    net_data = net(netfile, uniprot_data, meta_data)
    net_data.to_df().to_sql('NETWORK', con, index=False)

    # Generate Positive Network table
    if posnetfile != None:
        click.echo("Info: Parsing positive network file %s." % posnetfile)
        posnet_data = net(posnetfile, uniprot_data, meta_data)
        posnet_data.to_df().to_sql('POSITIVE_NETWORK', con, index=False)
    else:
        posnet_data = None

    # Generate Negative Network table
    if negnetfile != None:
        click.echo("Info: Parsing negative network file %s." % negnetfile)
        negnet_data = net(negnetfile, uniprot_data, meta_data)
        negnet_data.to_df().to_sql('NEGATIVE_NETWORK', con, index=False)
    else:
        negnet_data = None

    # Generate interaction query data
    click.echo("Info: Generating interaction query data.")
    query_data = query(net_data, posnet_data, negnet_data, meta_data.protein_meta, min_interaction_confidence, interaction_confidence_bins, interaction_confidence_quantile, decoy_oversample, decoy_subsample) # decoy_exclude
    query_data.to_df().to_sql('QUERY', con, index=False)

    # Remove any entries that are not necessary (proteins not covered by LC-MS/MS data)
    con.execute('DELETE FROM PROTEIN WHERE protein_id NOT IN (SELECT DISTINCT(protein_id) as protein_id FROM QUANTIFICATION);')
    con.execute('DELETE FROM NETWORK WHERE bait_id NOT IN (SELECT DISTINCT(protein_id) as protein_id FROM QUANTIFICATION) OR prey_id NOT IN (SELECT DISTINCT(protein_id) as protein_id FROM QUANTIFICATION);')
    con.execute('DELETE FROM SEC WHERE run_id NOT IN (SELECT DISTINCT(run_id) as run_id FROM QUANTIFICATION);')
    con.execute('DELETE FROM QUERY WHERE bait_id NOT IN (SELECT DISTINCT(protein_id) as protein_id FROM QUANTIFICATION) OR prey_id NOT IN (SELECT DISTINCT(protein_id) as protein_id FROM QUANTIFICATION);')

    # Add indices
    con.execute('CREATE INDEX IF NOT EXISTS idx_protein_protein_id ON PROTEIN (protein_id);')
    con.execute('CREATE INDEX IF NOT EXISTS idx_network_bait_id ON NETWORK (bait_id);')
    con.execute('CREATE INDEX IF NOT EXISTS idx_network_prey_id ON NETWORK (prey_id);')
    con.execute('CREATE INDEX IF NOT EXISTS idx_network_bait_id_prey_id ON NETWORK (bait_id, prey_id);')
    con.execute('CREATE INDEX IF NOT EXISTS idx_quantification_run_id ON QUANTIFICATION (run_id);')
    con.execute('CREATE INDEX IF NOT EXISTS idx_quantification_protein_id ON QUANTIFICATION (protein_id);')
    con.execute('CREATE INDEX IF NOT EXISTS idx_quantification_peptide_id ON QUANTIFICATION (peptide_id);')
    con.execute('CREATE INDEX IF NOT EXISTS idx_peptide_meta_peptide_id ON PEPTIDE_META (peptide_id);')
    con.execute('CREATE INDEX IF NOT EXISTS idx_protein_meta_protein_id ON PROTEIN_META (protein_id);')
    con.execute('CREATE INDEX IF NOT EXISTS idx_query_bait_id ON QUERY (bait_id);')
    con.execute('CREATE INDEX IF NOT EXISTS idx_query_prey_id ON QUERY (prey_id);')
    con.execute('CREATE INDEX IF NOT EXISTS idx_query_bait_id_prey_id ON QUERY (bait_id, prey_id);')
    # con.execute('VACUUM;')

    con.commit()

    # Close connection to file
    con.close()

    click.echo("Info: Data successfully preprocessed and stored in %s." % outfile)

# SECAT score features
@cli.command()
@click.option('--in', 'infile', required=True, type=click.Path(exists=True), help='Input SECAT file.')
@click.option('--out', 'outfile', required=False, type=click.Path(exists=False), help='Output SECAT file.')
@click.option('--monomer_threshold_factor', 'monomer_threshold_factor', default=2.0, show_default=True, type=float, help='Factor threshold to consider a feature a complex rather than a monomer.')
@click.option('--minimum_peptides', 'minimum_peptides', default=1, show_default=True, type=int, help='Minimum number of peptides required to score an interaction.')
@click.option('--maximum_peptides', 'maximum_peptides', default=3, show_default=True, type=int, help='Maximum number of peptides used to score an interaction.')
@click.option('--peakpicking', default='none', show_default=True, type=click.Choice(['none', 'detrend_zero', 'detrend_drop', 'localmax_conditions', 'localmax_replicates']), help='Either "none", "detrend_zero", "detrend_drop", "localmax_conditions" or "localmax_replicates"; the method for peakpicking of the peptide chromatograms. detrend_drop averages over all fractions with peptides; detrend_zero averages over all fractions (less agressive). localmax_conditions averages peak-picking over replicates of the same conditions; localmax_replicates conducts peak-picking for all samples separately.')
@click.option('--chunck_size', 'chunck_size', default=50000, show_default=True, type=int, help='Chunck size for processing.')
@click.option('--threads', default=1, show_default=True, type=int, help='Number of threads used for parallel processing. -1 means all available CPUs.', callback=transform_threads)
def score(infile, outfile, monomer_threshold_factor, minimum_peptides, maximum_peptides, peakpicking, chunck_size, threads):
    """
    Score interaction features in SEC data.
    """

    # Define outfile
    if outfile is None:
        outfile = infile
    else:
        # TODO: Consider replacing this with subprocess.call(["cp", "infile", "outfile"]) for speed improvement
        copyfile(infile, outfile)
        outfile = outfile

    # Find monomer thresholds
    click.echo("Info: Detect monomers.")
    monomer_data = monomer(outfile, monomer_threshold_factor)

    con = connect(outfile)
    monomer_data.df.to_sql('MONOMER', con, index=False, if_exists='replace')
    con.close()

    # Signal processing
    click.echo("Info: Signal processing.")

    # Drop features if they already exist
    con = connect(outfile)
    c = con.cursor()
    c.execute('DROP TABLE IF EXISTS FEATURE;')
    con.close()

    scoring(outfile, chunck_size, threads, minimum_peptides, maximum_peptides, peakpicking)
    # cProfile.runctx('scoring(outfile, chunck_size, minimum_peptides, maximum_peptides, peakpicking)', globals(), locals = {'outfile': outfile, 'chunck_size': chunck_size, 'minimum_peptides': minimum_peptides, 'maximum_peptides': maximum_peptides, 'peakpicking': peakpicking}, filename="score_performance.cprof")

# SECAT learn features
@cli.command()
@click.option('--in', 'infile', required=True, type=click.Path(exists=True), help='Input SECAT file.')
@click.option('--out', 'outfile', required=False, type=click.Path(exists=False), help='Output SECAT file.')
@click.option('--apply_model', 'apply_model', required=False, type=click.Path(exists=False), help='Apply pretrained SECAT model')
# Prefiltering
@click.option('--minimum_abundance_ratio', 'minimum_abundance_ratio', default=0.1, show_default=True, type=float, help='Minimum abundance ratio required to score an interaction.')
@click.option('--maximum_sec_shift', 'maximum_sec_shift', default=10, show_default=True, type=float, help='Maximum lag in SEC units between interactions and subunits.')
@click.option('--cb_decoys/--no-cb_decoys', default=False, show_default=True, help='Use only decoys from same confidence bin instead of full set for learning.')
# Semi-supervised learning
@click.option('--xeval_fraction', default=0.8, show_default=True, type=float, help='Data fraction used for cross-validation of semi-supervised learning step.')
@click.option('--xeval_num_iter', default=3, show_default=True, type=int, help='Number of iterations for cross-validation of semi-supervised learning step.')
@click.option('--ss_initial_fdr', default=0.1, show_default=True, type=float, help='Initial FDR cutoff for best scoring targets.')
@click.option('--ss_iteration_fdr', default=0.05, show_default=True, type=float, help='Iteration FDR cutoff for best scoring targets.')
@click.option('--ss_num_iter', default=10, show_default=True, type=int, help='Number of iterations for semi-supervised learning step.')
@click.option('--xgb_autotune/--no-xgb_autotune', default=False, show_default=True, help='Autotune hyperparameters after semi-supervised learning.')
# Statistics
@click.option('--parametric/--no-parametric', default=False, show_default=True, help='Do parametric estimation of p-values.')
@click.option('--pfdr/--no-pfdr', default=False, show_default=True, help='Compute positive false discovery rate (pFDR) instead of FDR.')
@click.option('--pi0_lambda', default=[0.01,0.5,0.01], show_default=True, type=(float, float, float), help='Use non-parametric estimation of p-values. Either use <START END STEPS>, e.g. 0.1, 1.0, 0.1 or set to fixed value, e.g. 0.4, 0, 0.', callback=transform_pi0_lambda)
@click.option('--pi0_method', default='bootstrap', show_default=True, type=click.Choice(['smoother', 'bootstrap']), help='Either "smoother" or "bootstrap"; the method for automatically choosing tuning parameter in the estimation of pi_0, the proportion of true null hypotheses.')
@click.option('--pi0_smooth_df', default=3, show_default=True, type=int, help='Number of degrees-of-freedom to use when estimating pi_0 with a smoother.')
@click.option('--pi0_smooth_log_pi0/--no-pi0_smooth_log_pi0', default=False, show_default=True, help='If True and pi0_method = "smoother", pi0 will be estimated by applying a smoother to a scatterplot of log(pi0) estimates against the tuning parameter lambda.')
@click.option('--lfdr_truncate/--no-lfdr_truncate', show_default=True, default=True, help='If True, local FDR values >1 are set to 1.')
@click.option('--lfdr_monotone/--no-lfdr_monotone', show_default=True, default=True, help='If True, local FDR values are non-decreasing with increasing p-values.')
@click.option('--lfdr_transformation', default='probit', show_default=True, type=click.Choice(['probit', 'logit']), help='Either a "probit" or "logit" transformation is applied to the p-values so that a local FDR estimate can be formed that does not involve edge effects of the [0,1] interval in which the p-values lie.')
@click.option('--lfdr_adj', default=1.5, show_default=True, type=float, help='Numeric value that is applied as a multiple of the smoothing bandwidth used in the density estimation.')
@click.option('--lfdr_eps', default=power(10.0,-8), show_default=True, type=float, help='Numeric value that is threshold for the tails of the empirical p-value distribution.')
#other
@click.option('--plot_reports/--no-plot_reports', default=False, show_default=True, help='Plot reports for all confidence bins.')
@click.option('--threads', default=1, show_default=True, type=int, help='Number of threads used for parallel processing. -1 means all available CPUs.', callback=transform_threads)
@click.option('--test/--no-test', default=False, show_default=True, help='Run in test mode with fixed seed to ensure reproducibility.')
@click.option('--export_tables/--no-export_tables', default=False, show_default=True, help='Saves two csv tables. One for the interations in used for modeling interactions, and another of the target interactions. Including all scores.')
def learn(infile, outfile, apply_model, minimum_abundance_ratio, maximum_sec_shift, cb_decoys, xeval_fraction, xeval_num_iter, ss_initial_fdr, ss_iteration_fdr, ss_num_iter, xgb_autotune, parametric, pfdr, pi0_lambda, pi0_method, pi0_smooth_df, pi0_smooth_log_pi0, lfdr_truncate, lfdr_monotone, lfdr_transformation, lfdr_adj, lfdr_eps, plot_reports, threads, test, export_tables):
    """
    Learn true/false interaction features in SEC data.
    """

    # Define outfile
    if outfile is None:
        outfile = infile
    else:
        # TODO: Consider replacing this with subprocess.call(["cp", "infile", "outfile"]) for speed improvement
        copyfile(infile, outfile)
        outfile = outfile

    # Run PyProphet training
    click.echo("Info: Running PyProphet.")

    # Drop feature scores if they already exist
    con = connect(outfile)
    c = con.cursor()
    c.execute('DROP TABLE IF EXISTS FEATURE_SCORED;')
    con.close()

    pyprophet(outfile, apply_model, minimum_abundance_ratio, maximum_sec_shift, cb_decoys, xeval_fraction, xeval_num_iter, ss_initial_fdr, ss_iteration_fdr, ss_num_iter, xgb_autotune, parametric, pfdr, pi0_lambda, pi0_method, pi0_smooth_df, pi0_smooth_log_pi0, lfdr_truncate, lfdr_monotone, lfdr_transformation, lfdr_adj, lfdr_eps, plot_reports, threads, test, export_tables)

    # Combine all replicates
    click.echo("Info: Combine evidence across replicate runs.")

    combined_data = combine(outfile, pi0_lambda, pi0_method, pi0_smooth_df, pi0_smooth_log_pi0, pfdr)

    con = connect(outfile)
    if export_tables != False:
        network_interaction_name = path.splitext(infile)[0] + "_net_int_scored.csv"
        combined_data.df.to_csv(network_interaction_name, index=False)
    combined_data.df.to_sql('FEATURE_SCORED_COMBINED', con, index=False, if_exists='replace')
    con.close()

# SECAT quantify features
@cli.command()
@click.option('--in', 'infile', required=True, type=click.Path(exists=True), help='Input SECAT file.')
@click.option('--out', 'outfile', required=False, type=click.Path(exists=False), help='Output SECAT file.')
@click.option('--control_condition', default="center", type=str, help='Specify control condition identifier. Setting this parameter to "center" will compare all conditions against all and use the mean as reference for quantification.')
@click.option('--paired/--no-paired', default=False, show_default=True, help='Whether replicates should be paired, e.g. replicates 1 of conditions A & B were measured with heavy and light SILAC labels as part of the same runs.')
@click.option('--maximum_interaction_qvalue', default=0.05, show_default=True, type=float, help='Maximum q-value to consider interactions for quantification.')
@click.option('--min_abs_log2fx', default=1.0, show_default=True, type=float, help='Minimum absolute log2 fold-change for integrated nodes.')
@click.option('--minimum_peptides', 'minimum_peptides', default=1, show_default=True, type=int, help='Minimum number of peptides required to quantify an interaction.')
@click.option('--maximum_peptides', 'maximum_peptides', default=3, show_default=True, type=int, help='Maximum number of peptides used to quantify an interaction.')
@click.option('--missing_peptides', 'missing_peptides', default="zero", type=str, help='Whether missing peptide abundances should be set to 0 ("zero") or dropped ("drop") for fold change computation.')
@click.option('--peptide_log2fx/--no-peptide_log2fx', default=True, show_default=True, help='Whether peptide-level log2fx should be computed instead of protein-level. Protein-level is more robust if measured peptides are variable between conditions or replicates.')
@click.option('--threads', default=1, show_default=True, type=int, help='Number of threads used for parallel processing. -1 means all available CPUs.', callback=transform_threads)
def quantify(infile, outfile, control_condition, paired, maximum_interaction_qvalue, min_abs_log2fx, minimum_peptides, maximum_peptides, missing_peptides, peptide_log2fx, threads):
    """
    Quantify protein and interaction features in SEC data.
    """

    # Define outfile
    if outfile is None:
        outfile = infile
    else:
        # TODO: Consider replacing this with subprocess.call(["cp", "infile", "outfile"]) for speed improvement
        copyfile(infile, outfile)
        outfile = outfile

    click.echo("Info: Prepare quantitative matrices.")
    qm = quantitative_matrix(outfile, maximum_interaction_qvalue, minimum_peptides, maximum_peptides)

    con = connect(outfile)
    qm.monomer_peptide.to_sql('MONOMER_QM', con, index=False, if_exists='replace')
    qm.complex_peptide.to_sql('COMPLEX_QM', con, index=False, if_exists='replace')
    con.close()

    click.echo("Info: Assess differential features.")
    et = enrichment_test(outfile, control_condition, paired, min_abs_log2fx, missing_peptides, peptide_log2fx, threads)


    con = connect(outfile)
    et.edge.to_sql('EDGE', con, index=False, if_exists='replace')
    et.edge_level.to_sql('EDGE_LEVEL', con, index=False, if_exists='replace')
    et.node.to_sql('NODE', con, index=False, if_exists='replace')
    et.node_level.to_sql('NODE_LEVEL', con, index=False, if_exists='replace')
    et.protein_level.to_sql('PROTEIN_LEVEL', con, index=False, if_exists='replace')

    con.close()

# SECAT export features
@cli.command()
@click.option('--in', 'infile', required=True, type=click.Path(exists=True), help='Input SECAT file.')
@click.option('--level', default='bait', show_default=True, type=click.Choice(['bait', 'interaction']), help='Export either all interactions of bait proteins or individual interactions')
@click.option('--id', required=False, type=str, help='Export specific UniProt bait_id (Q10000) or interaction_id (Q10000_P10000)')
@click.option('--max_qvalue', default=0.01, show_default=True, type=float, help='Maximum q-value to export baits or interactions.')
@click.option('--min_abs_log2fx', default=1.0, show_default=True, type=float, help='Minimum absolute log2 fold-change for integrated nodes.')
@click.option('--mode', default='quantitative', show_default=True, type=click.Choice(['quantitative', 'detection']), help='Select mode to order interaction exported tables by.')
@click.option('--combined/--no-combined', default=False, show_default=True, help='Select interactions and baits according to combined q-values.')
@click.option('--peptide_rank', default=6, show_default=True, type=int, help='Number of most intense peptides to export.')
@click.option('--extra', default=False, show_default=True, type=bool, help='Whether or not to export raw csv files of data used in `secat plot`')
def export(infile, level, id, max_qvalue, min_abs_log2fx, mode, combined, peptide_rank, extras):
    """
    Export SECAT results.
    """
    export_tables(infile, level, id, max_qvalue, min_abs_log2fx, mode, combined, peptide_rank, extras)    


# SECAT plot chromatograms
@cli.command()
@click.option('--in', 'infile', required=True, type=click.Path(exists=True), help='Input SECAT file.')
@click.option('--level', default='bait', show_default=True, type=click.Choice(['bait', 'interaction']), help='Plot either all interactions of bait proteins or individual interactions')
@click.option('--id', required=False, type=str, help='Plot specific UniProt bait_id (Q10000) or interaction_id (Q10000_P10000)')
@click.option('--max_qvalue', default=0.01, show_default=True, type=float, help='Maximum q-value to plot baits or interactions.')
@click.option('--min_abs_log2fx', default=1.0, show_default=True, type=float, help='Minimum absolute log2 fold-change for integrated nodes.')
@click.option('--mode', default='quantitative', show_default=True, type=click.Choice(['quantitative', 'detection']), help='Select mode to order interaction plots by.')
@click.option('--combined/--no-combined', default=False, show_default=True, help='Select interactions and baits according to combined q-values.')
@click.option('--peptide_rank', default=6, show_default=True, type=int, help='Number of most intense peptides to plot.')
def plot(infile, level, id, max_qvalue, min_abs_log2fx, mode, combined, peptide_rank):
    """
    Plot SECAT results
    """

    pf = plot_features(infile, level, id, max_qvalue, min_abs_log2fx, mode, combined, peptide_rank)

# SECAT print statistics
@cli.command()
@click.option('--in', 'infile', required=True, type=click.Path(exists=True), help='Input SECAT file.')
@click.option('--min_abs_log2fx', default=1.0, show_default=True, type=float, help='Minimum absolute log2 fold-change for integrated nodes.')
def statistics(infile, min_abs_log2fx):
    """
    Print SECAT statistics
    """

    con = connect(infile)

    if check_sqlite_table(con, 'QUANTIFICATION') and check_sqlite_table(con, 'SEC') and check_sqlite_table(con, 'PROTEIN'):
        click.echo("Protein information")
        click.echo(10*"-")
        pepprot = read_sql('SELECT * FROM QUANTIFICATION INNER JOIN SEC ON QUANTIFICATION.run_id = SEC.run_id WHERE protein_id IN (SELECT DISTINCT protein_id FROM PROTEIN);' , con)

        click.echo("Total proteins: %s" % len(pepprot['protein_id'].drop_duplicates()))
        click.echo("Total proteins per run:")
        click.echo(pepprot.groupby(['condition_id', 'replicate_id'])['protein_id'].nunique())
        click.echo(10*"-")

        click.echo("Total peptides: %s" % len(pepprot[['peptide_id']].drop_duplicates()))
        click.echo("Total peptides per run:")
        click.echo(pepprot.groupby(['condition_id', 'replicate_id'])['peptide_id'].nunique())
        click.echo(10*"=")

    if check_sqlite_table(con, 'FEATURE_SCORED_COMBINED'):
        click.echo("PPI Information")
        click.echo(10*"-")
        intact_combined = read_sql('SELECT *, bait_id || "_" || prey_id AS interaction_id FROM FEATURE_SCORED_COMBINED WHERE decoy==0;' , con)
        intact = read_sql('SELECT *, bait_id || "_" || prey_id AS interaction_id FROM FEATURE_SCORED WHERE decoy==0;' , con)

        click.echo("Total interactions (q-value<0.01): %s" % intact_combined[intact_combined['qvalue']<0.01]['interaction_id'].nunique())
        click.echo("Total interactions per run (q-value<0.01):")
        click.echo(intact[intact['qvalue']<0.01].groupby(['condition_id', 'replicate_id'])['interaction_id'].nunique())
        click.echo(10*"-")

        click.echo("Total interactions (q-value<0.05): %s" % intact_combined[intact_combined['qvalue']<0.05]['interaction_id'].nunique())
        click.echo("Total interactions per run (q-value<0.05):")
        click.echo(intact[intact['qvalue']<0.05].groupby(['condition_id', 'replicate_id'])['interaction_id'].nunique())
        click.echo(10*"-")

        click.echo("Total interactions (q-value<0.1): %s" % intact_combined[intact_combined['qvalue']<0.1]['interaction_id'].nunique())
        click.echo("Total interactions per run (q-value<0.1):")
        click.echo(intact[intact['qvalue']<0.1].groupby(['condition_id', 'replicate_id'])['interaction_id'].nunique())
        click.echo(10*"=")

    if check_sqlite_table(con, 'PROTEIN_LEVEL'):
        click.echo("Quantitative Information (min_abs_log2fx > %s)" % min_abs_log2fx)
        click.echo(10*"-")

        df_node_level = read_sql('SELECT * FROM NODE_LEVEL;', con)

        df_node_level_filtered = df_node_level[df_node_level['abs_log2fx'] > min_abs_log2fx]
        df_node = df_node_level_filtered.sort_values('pvalue_adjusted').groupby(['condition_1','condition_2','bait_id']).head(1).reset_index()

        click.echo("Info: Total dysregulated proteins detected:")
        click.echo("%s (at FDR < 0.01)" % (df_node[df_node['pvalue_adjusted'] < 0.01][['bait_id']].drop_duplicates().shape[0]))
        click.echo("%s (at FDR < 0.05)" % (df_node[df_node['pvalue_adjusted'] < 0.05][['bait_id']].drop_duplicates().shape[0]))
        click.echo("%s (at FDR < 0.1)" % (df_node[df_node['pvalue_adjusted'] < 0.1][['bait_id']].drop_duplicates().shape[0]))
        click.echo("%s (unfiltered)" % (df_node[['bait_id']].drop_duplicates().shape[0]))

        for level in df_node_level_filtered['level'].unique():
            click.echo("Info: Dysregulated (%s-mode) proteins detected:" % (level))
            click.echo("%s (at FDR < 0.01)" % (df_node_level_filtered[(df_node_level_filtered['level'] == level) & (df_node_level_filtered['pvalue_adjusted'] < 0.01)][['bait_id']].drop_duplicates().shape[0]))
            click.echo("%s (at FDR < 0.05)" % (df_node_level_filtered[(df_node_level_filtered['level'] == level) & (df_node_level_filtered['pvalue_adjusted'] < 0.05)][['bait_id']].drop_duplicates().shape[0]))
            click.echo("%s (at FDR < 0.1)" % (df_node_level_filtered[(df_node_level_filtered['level'] == level) & (df_node_level_filtered['pvalue_adjusted'] < 0.1)][['bait_id']].drop_duplicates().shape[0]))
            click.echo("%s (unfiltered)" % (df_node_level_filtered[(df_node_level_filtered['level'] == level)][['bait_id']].drop_duplicates().shape[0]))
            
        click.echo(10*"=")



import pandas as pd
import numpy as np
import click
import pdb
import sqlite3
import sys
from decoupler.method_viper import run_viper
import os

from .EmpiricalBrownsMethod import EmpiricalBrownsMethod
import itertools

from scipy.stats import ttest_ind, ttest_rel

from statsmodels.stats.multitest import multipletests

class quantitative_matrix:
    def __init__(self, outfile, maximum_interaction_qvalue, minimum_peptides, maximum_peptides):
        self.outfile = outfile
        self.maximum_interaction_qvalue = maximum_interaction_qvalue
        self.minimum_peptides = minimum_peptides
        self.maximum_peptides = maximum_peptides

        self.interactions, self.detections, self.chromatograms, self.peaks = self.read()
        self.monomer_peptide = self.quantify_monomers()
        self.complex_peptide = self.quantify_complexes()

    def read(self):
        con = sqlite3.connect(self.outfile)

        click.echo("Getting interations table")
        interactions = pd.read_sql(
            f"""
            SELECT DISTINCT bait_id, prey_id 
            FROM FEATURE_SCORED_COMBINED 
            WHERE qvalue <= {self.maximum_interaction_qvalue} 
            AND bait_id != prey_id 
            AND decoy == 0;
            """, 
            con
        )

        click.echo("Getting detections table")
        detections = pd.read_sql(
            f"""
            SELECT DISTINCT condition_id, replicate_id, FEATURE_SCORED.bait_id, FEATURE_SCORED.prey_id 
            FROM FEATURE_SCORED 
            INNER JOIN (
                SELECT DISTINCT bait_id, prey_id 
                FROM FEATURE_SCORED_COMBINED 
                WHERE qvalue <= {self.maximum_interaction_qvalue} 
                AND bait_id != prey_id 
                AND decoy == 0
            ) AS FEATURE_SCORED_COMBINED 
            ON FEATURE_SCORED.bait_id = FEATURE_SCORED_COMBINED.bait_id 
            AND FEATURE_SCORED.prey_id = FEATURE_SCORED_COMBINED.prey_id;
            """, 
            con
        )

        click.echo("Getting chromatograms table")
        chromatograms = pd.read_sql(
            f"""
            SELECT 
                SEC.condition_id, 
                SEC.replicate_id, 
                SEC.sec_id, 
                QUANTIFICATION.protein_id, 
                QUANTIFICATION.peptide_id, 
                peptide_intensity, 
                MONOMER.sec_id AS monomer_sec_id 
            FROM QUANTIFICATION 
            INNER JOIN PROTEIN_META ON QUANTIFICATION.protein_id = PROTEIN_META.protein_id 
            INNER JOIN PEPTIDE_META ON QUANTIFICATION.peptide_id = PEPTIDE_META.peptide_id 
            INNER JOIN SEC ON QUANTIFICATION.RUN_ID = SEC.RUN_ID 
            INNER JOIN MONOMER ON QUANTIFICATION.protein_id = MONOMER.protein_id AND SEC.condition_id = MONOMER.condition_id AND SEC.replicate_id = MONOMER.replicate_id 
            WHERE peptide_count >= {self.minimum_peptides} 
            AND peptide_rank <= {self.maximum_peptides};
            """, 
            con
        )

        # TODO: Consider replacing pd.read_sql with https://github.com/sfu-db/connector-x to improve read speeds and utilize concurrency
        click.echo("Getting peaks table")
        peaks = pd.read_sql('SELECT * FROM PROTEIN_PEAKS;', con)

        con.close()

        return interactions, detections, chromatograms, peaks

    def quantify_monomers(self):
        def sec_summarize(df):
            def aggregate(x):
                peptide_ix = (x['peptide_intensity']-x['peptide_intensity'].max()).abs().argsort()[:self.maximum_peptides]
                return pd.DataFrame({'peptide_id': x.iloc[peptide_ix]['peptide_id'], 'peptide_intensity': x.iloc[peptide_ix]['peptide_intensity'], 'total_peptide_intensity': x.iloc[peptide_ix]['total_peptide_intensity']})

            # Summarize total peptide intensities
            peptide_total = df.groupby(['peptide_id'])['peptide_intensity'].sum().reset_index()
            peptide_total.columns = ['peptide_id','total_peptide_intensity']

            # Summarize monomer peptide intensities
            if df[df['sec_id'] >= df['monomer_sec_id']].shape[0] > 0:
                peptide_mono = df[df['sec_id'] >= df['monomer_sec_id']].groupby(['peptide_id'])['peptide_intensity'].sum().reset_index()
                peptide_mono.columns = ['peptide_id','peptide_intensity']
            else:
                peptide_mono = pd.DataFrame({'peptide_id': df['peptide_id'].unique(), 'peptide_intensity': 0})

            peptide = pd.merge(peptide_mono, peptide_total, on='peptide_id')

            # Ensure that minimum peptides are present
            if peptide.shape[0] >= self.minimum_peptides:
                # Select representative closest to max and aggregate
                peptide = aggregate(peptide)

                return peptide

        def peptide_summarize(df):
            # Aggregate to peptide level
            peptide = df[['condition_id','replicate_id','bait_id','prey_id','is_bait','peptide_id']].copy()
            peptide['monomer_abundance'] = np.log2(df['peptide_intensity'].values+1)
            peptide['assembled_abundance'] = np.log2(df['total_peptide_intensity'].values-df['peptide_intensity'].values+1)
            peptide['total_abundance'] = np.log2(df['total_peptide_intensity'].values+1)

            return peptide

        # Quantify monomers
        monomers = self.chromatograms.copy()
        monomers['bait_id'] = monomers['protein_id']
        monomers['prey_id'] = monomers['protein_id']
        monomers['is_bait'] = True

        monomers_sec = monomers.groupby(['condition_id','replicate_id','bait_id','prey_id','is_bait']).apply(sec_summarize).reset_index(level=['condition_id','replicate_id','bait_id','prey_id','is_bait'])
        monomers_peptides = peptide_summarize(monomers_sec)

        return monomers_peptides

    def quantify_complexes(self):
        def sec_summarize(df):
            def aggregate(x):
                peptide_ix = (x['peptide_intensity'] - x['peptide_intensity'].max()).abs().argsort().iloc[:self.maximum_peptides]
                return pd.DataFrame({'peptide_id': x.iloc[peptide_ix]['peptide_id'], 'peptide_intensity': x.iloc[peptide_ix]['peptide_intensity']})

            # Remove monomer fractions for complex-centric quantification
            df = df[df['sec_id'] < df['monomer_sec_id']]
            # Find SEC intersections
            bait_sec = df[df['is_bait']]['sec_id'].unique()
            prey_sec = df[~df['is_bait']]['sec_id'].unique()
            intersection = pd.DataFrame({'sec_id': list(set(bait_sec) & set(prey_sec))})
            # There needs to be at least one fraction where peptides from both proteins are measured.
            if intersection.shape[0] > 0:
                # Summarize intersection peptide intensities
                df_is = pd.merge(df, intersection, on='sec_id')
                peptide_is = df_is.groupby(['peptide_id'])['peptide_intensity'].sum().reset_index()
                peptide_is.columns = ['peptide_id','peptide_intensity']

                peptide = pd.merge(df[['condition_id','replicate_id','bait_id','prey_id','is_bait','peptide_id']].drop_duplicates(), peptide_is, on='peptide_id')
                # Ensure that minimum peptides are present for both interactors for quantification.
                if peptide[peptide['is_bait']].shape[0] >= self.minimum_peptides and peptide[~peptide['is_bait']].shape[0] >= self.minimum_peptides:
                    # Select representative closest to max and aggregate
                    peptide = peptide.groupby(['is_bait'], group_keys=True).apply(aggregate).reset_index(level=['is_bait'])
                    
                    return peptide

        def peptide_summarize(df):
            # Aggregate to peptide level
            peptide = df[['condition_id','replicate_id','bait_id','prey_id','is_bait','peptide_id']].copy()
            peptide['interactor_abundance'] = np.log2(df['peptide_intensity']+1)

            return peptide

        # Restrict chromatographic data to selected peaks only
        chromatograms = pd.merge(self.chromatograms, self.peaks, on=['condition_id','replicate_id','protein_id','sec_id'])

        # Quantify interactions
        baits = pd.merge(self.interactions, chromatograms, left_on=['bait_id'], right_on=['protein_id']).drop(columns=['protein_id'])
        baits['is_bait'] = True

        preys = pd.merge(self.interactions, chromatograms, left_on=['prey_id'], right_on=['protein_id']).drop(columns=['protein_id'])
        preys['is_bait'] = False

        complexes = pd.concat([baits, preys]).reset_index()
        complexes_sec = complexes.groupby(['condition_id','replicate_id','bait_id','prey_id']).apply(sec_summarize).reset_index(level=['condition_id','replicate_id','bait_id','prey_id'])
        complexes_peptides = peptide_summarize(complexes_sec)

        return complexes_peptides

class enrichment_test:
    def __init__(self, outfile, control_condition, paired, min_abs_log2fx, missing_peptides, peptide_log2fx, threads):
        self.outfile = outfile
        self.control_condition = control_condition
        self.paired = paired
        self.min_abs_log2fx = min_abs_log2fx
        self.missing_peptides = missing_peptides
        self.peptide_log2fx = peptide_log2fx
        self.threads = threads
        self.levels = ['interactor_abundance','complex_abundance','interactor_ratio','monomer_abundance','assembled_abundance','total_abundance']
        self.comparisons = self.contrast()

        self.monomer_qm, self.complex_qm = self.read()

        self.tests = self.compare()
        self.edge_level, self.edge, self.node_level, self.node, self.protein_level = self.integrate()

    def contrast(self):
        con = sqlite3.connect(self.outfile)
        conditions = pd.read_sql('SELECT DISTINCT condition_id FROM SEC;' , con)['condition_id'].values.tolist()

        con.close()

        if len(conditions) < 2:
            sys.exit("Error: Your experimental design is not supported. At least two conditions are necessary for differential analysis.")

        comparisons = []
        # prepare single-sample comparisons
        if self.control_condition in conditions:
            conditions.remove(self.control_condition)
            for condition in conditions:
                comparisons.append([condition, self.control_condition])
        # prepare multi-sample comparisons
        elif self.control_condition == 'center':
            comparisons = list(itertools.combinations(conditions, 2))
        else:
            sys.exit("Error: Specify correct control condition identifier as reference or use 'center' to compare all against all.")

        return comparisons

    def read(self):
        con = sqlite3.connect(self.outfile)

        monomer_qm = pd.read_sql('SELECT * FROM MONOMER_QM;' , con)
        complex_qm = pd.read_sql('SELECT * FROM COMPLEX_QM;' , con)

        con.close()

        return monomer_qm, complex_qm

    def compare(self):
        # TODO: Add tqdm progress bar to better visualize compare() progress
        dfs = []
        for level in self.levels:
            for state in [self.monomer_qm, self.complex_qm]:
                if level in state.columns or (level in ['complex_abundance', 'interactor_ratio'] and 'interactor_abundance' in state.columns):
                    if level in state.columns:
                        dat = state[state[level] > 0].copy()
                    else:
                        dat = state[state['interactor_abundance'] > 0].copy()
                        dat = dat.rename(index=str, columns={"interactor_abundance": level})

                    dat['query_id'] = dat['bait_id'] + '_' + dat['prey_id']
                    dat['query_peptide_id'] = dat['bait_id'] + '_' + dat['prey_id'] + '_' + dat['peptide_id']
                    dat['quantification_id'] = 'viper_' + dat['condition_id'] + '_' + dat['replicate_id']
                    dat['run_id'] = dat['condition_id'] + '_' + dat['replicate_id']
                    qm_ids = dat[['quantification_id','condition_id','replicate_id']].drop_duplicates()

                    if self.missing_peptides == 'drop':
                        peptide_fill_value = np.nan
                    elif self.missing_peptides == 'zero':
                        peptide_fill_value = 0
                    else:
                        sys.exit("Error: Invalid parameter for 'missing_peptides' selected.")

                    # Generate matrix for fold-change
                    quant_mx = dat.pivot_table(index=['query_id','is_bait','query_peptide_id'], columns='quantification_id', values=level, fill_value=peptide_fill_value)

                    # Generate matrix for ratio-change
                    ratio_mx = dat.pivot_table(index=['query_id','is_bait','query_peptide_id'], columns='quantification_id', values=level, fill_value=peptide_fill_value)

                    # Generate matrix for VIPER
                    data_mx = dat.pivot_table(index='query_peptide_id', columns='quantification_id', values=level, fill_value=0)

                    net = pd.DataFrame(columns=["source", "target", "weight"])
                    
                    # Generate subunit set for VIPER
                    if level == 'complex_abundance':
                        # Complex abundance testing combines bait and prey peptides into a single regulon with positive tfmode sign
                        query_set = dat[['query_id','is_bait','query_peptide_id']].copy()
                        query_set['query_id'] = query_set['query_id'] + "+1"

                        net['source'] = query_set['query_id']
                        net['target'] = query_set['query_peptide_id']
                        net['weight'] = [1 for _ in range(query_set.shape[0])]
                        net = net.groupby(['source', 'target']).sum('weight').reset_index()
                    elif level == 'interactor_ratio':
                        # Complex stoichiometry testing combines bait and prey peptides into a single regulon but with different tfmode signs
                        query_set = dat[['query_id','is_bait','query_peptide_id']].copy()
                        query_set.loc[query_set['is_bait']==0,'is_bait'] = -1
                        query_set['query_id'] = query_set['query_id'] + "+1"
                        filtered_query_set = query_set.groupby(['query_id']).apply(lambda x: x.drop_duplicates(['is_bait', 'query_peptide_id'])).reset_index(drop=True)

                        net['source'] = filtered_query_set['query_id']
                        net['target'] = filtered_query_set['query_peptide_id']
                        net['weight'] = filtered_query_set['is_bait'].apply(lambda x: 1 if x == True else -1)
                        net = net.groupby(['source', 'target']).sum('weight').reset_index() # Not sure if this is needed
                    else:
                        # All other modalities are assessed on protein-level, separately for bait and prey proteins
                        query_set = dat[['query_id','is_bait','query_peptide_id']].copy()
                        query_set['query_id'] = query_set['query_id'] + "+" + query_set['is_bait'].astype(int).astype(str)

                        net['source'] = query_set['query_id']
                        net['target'] = query_set['query_peptide_id']
                        net['weight'] = [1 for _ in range(query_set.shape[0])]
                        net = net.groupby(['source', 'target']).sum('weight').reset_index()

                    results = run_viper(data_mx.T, net, verbose=False, pleiotropy=False, min_n=1)
                    results = results[0].T.reset_index()
                    results['query_id'] = results['index']
                    results.drop('index', inplace=True, axis=1)

                    results[['query_id','is_bait']] = results['query_id'].str.split("+", expand=True)
                    results['is_bait'] = results['is_bait'].astype('int')
                    results['level'] = level

                    # Append reverse information for complex_abundance and interactor_ratio levels
                    if level in ['complex_abundance', 'interactor_ratio']:
                        results_rev = results.copy()
                        results_rev['is_bait'] = 0
                        results = pd.concat([results, results_rev])

                    for comparison in self.comparisons:
                        results['condition_1'] = comparison[0]
                        results['condition_2'] = comparison[1]

                        # Compute fold-change and absolute fold-change
                        quant_mx_avg = quant_mx.groupby(['query_id','is_bait','query_peptide_id']).apply(lambda x: pd.Series({'comparison_0': np.nanmean(np.exp2(x[qm_ids[qm_ids['condition_id']==comparison[0]]['quantification_id'].values].values)), 'comparison_1': np.nanmean(np.exp2(x[qm_ids[qm_ids['condition_id']==comparison[1]]['quantification_id'].values].values))})).reset_index(level=['query_id','is_bait','query_peptide_id'])

                        if self.peptide_log2fx:
                            quant_mx_log2fx = quant_mx_avg.groupby(['query_id','is_bait','query_peptide_id']).apply(lambda x: np.log2((x['comparison_0'])/(x['comparison_1']))).reset_index(level=['query_id','is_bait','query_peptide_id'])

                            quant_mx_log2fx_prot = quant_mx_log2fx.groupby(['query_id','is_bait']).mean(numeric_only=True).reset_index()
                            quant_mx_log2fx_prot.columns = ['query_id','is_bait','log2fx']
                            quant_mx_log2fx_prot['abs_log2fx'] = np.abs(quant_mx_log2fx_prot['log2fx'])
                        else:
                            quant_mx_avg_prot = quant_mx_avg.groupby(['query_id','is_bait'])[['comparison_0','comparison_1']].mean().reset_index()
                            quant_mx_log2fx_prot = quant_mx_avg_prot.groupby(['query_id','is_bait']).apply(lambda x: np.log2((x['comparison_0'])/(x['comparison_1']))).reset_index(level=['query_id','is_bait'])

                            quant_mx_log2fx_prot.columns = ['query_id','is_bait','log2fx']
                            quant_mx_log2fx_prot['abs_log2fx'] = np.abs(quant_mx_log2fx_prot['log2fx'])

                        results = pd.merge(results, quant_mx_log2fx_prot, on=['query_id','is_bait'], how='left')

                        # Compute interactor ratio
                        if level in ['complex_abundance', 'interactor_ratio']:
                            ratio_mx_prot = ratio_mx.groupby(['query_id','is_bait'])[[c for c in ratio_mx.columns if c.startswith("viper_")]].mean().reset_index()
                            ratio_mx_prot_ratio = ratio_mx_prot.groupby('query_id').apply(lambda x: (x.loc[x['is_bait']==0].squeeze()+1) / (x.loc[x['is_bait']==1].squeeze()+1)).reset_index(level='query_id')

                            ratio_change = ratio_mx_prot_ratio.groupby('query_id').apply(lambda x: np.mean(x[qm_ids[qm_ids['condition_id']==comparison[0]]['quantification_id'].values].values) / np.mean(x[qm_ids[qm_ids['condition_id']==comparison[1]]['quantification_id'].values].values)).reset_index(level='query_id')
                            ratio_change.columns = ['query_id','interactor_ratio']
                            ratio_change.loc[ratio_change['interactor_ratio'] > 1,'interactor_ratio'] = (1 / ratio_change.loc[ratio_change['interactor_ratio'] > 1,'interactor_ratio'])
                            results = pd.merge(results, ratio_change, on=['query_id'], how='left')
                        else:
                            results['interactor_ratio'] = np.nan

                        # Conduct statistical tests
                        # Paired analysis: For example replicates 1 of conditions A & B were measured by the same SILAC experiment
                        if self.paired:
                            results_pvalue = results.groupby(['query_id','is_bait','level']).apply(lambda x: pd.Series({"pvalue": ttest_rel(x[qm_ids[qm_ids['condition_id']==comparison[0]].sort_values(by=['quantification_id'])['quantification_id'].values].values[0], x[qm_ids[qm_ids['condition_id']==comparison[1]].sort_values(by=['quantification_id'])['quantification_id'].values].values[0])[1]})).reset_index()
                        # Treat samples as independent measurements, e.g. quantification by LFQ
                        else:
                            results_pvalue = results.groupby(['query_id','is_bait','level']).apply(lambda x: pd.Series({"pvalue": ttest_ind(x[qm_ids[qm_ids['condition_id']==comparison[0]]['quantification_id'].values].values[0], x[qm_ids[qm_ids['condition_id']==comparison[1]]['quantification_id'].values].values[0], equal_var=True)[1]})).reset_index()
                        results = pd.merge(results, results_pvalue, on=['query_id','is_bait','level'])

                        # Set p-value to 1.0 if invalid
                        results.loc[np.isnan(results['pvalue']),'pvalue'] = 1.0

                        # Append meta information
                        results = pd.merge(results, dat[['query_id','bait_id','prey_id']].drop_duplicates(), on='query_id')

                        dfs.append(results[['condition_1','condition_2','level','bait_id','prey_id','is_bait','log2fx','abs_log2fx','interactor_ratio','pvalue']+[c for c in results.columns if c.startswith("viper_")]])

        return pd.concat(dfs, ignore_index=True, sort=True).sort_values(by='pvalue', ascending=True, na_position='last')

    def integrate(self):
        def collapse(x):
            if x.shape[0] > 1:
                result = pd.Series({'num_interactors': x.shape[0], 'log2fx': np.mean(x['log2fx'].values), 'abs_log2fx': np.mean(x['abs_log2fx'].values), 'interactor_ratio': np.mean(x['interactor_ratio'].values), 'pvalue': EmpiricalBrownsMethod(x[[c for c in x.columns if c.startswith("viper_")]].values, x['pvalue'].values)})
            else:
                result = pd.Series({'num_interactors': x.shape[0], 'log2fx': x['log2fx'].values[0], 'abs_log2fx': x['abs_log2fx'].values[0], 'interactor_ratio': x['interactor_ratio'].values[0], 'pvalue': x['pvalue'].values[0]})
            return(result)

        def mtcorrect(x):
            x['pvalue_adjusted'] = multipletests(pvals=x['pvalue'].values, method="fdr_bh")[1]

            return(x)

        df_edge_level = self.tests[(self.tests['bait_id'] != self.tests['prey_id']) & self.tests['is_bait']]
        df_edge_level_rev = self.tests[(self.tests['bait_id'] != self.tests['prey_id']) & ~self.tests['is_bait']]
        df_edge_level_rev = df_edge_level_rev.rename(index=str, columns={"bait_id": "prey_id", "prey_id": "bait_id"})

        df_protein_level = self.tests[self.tests['bait_id'] == self.tests['prey_id']]
        df_edge_full = pd.concat([df_protein_level, df_edge_level, df_edge_level_rev], sort=False)
        df_edge_level = pd.concat([df_edge_level, df_edge_level_rev[df_edge_level_rev['level']=='interactor_abundance']], sort=False)
        df_node_level = df_edge_full.groupby(['condition_1', 'condition_2','level','bait_id'], group_keys=False).apply(collapse).reset_index()

        # Multi-testing correction and pooling
        df_protein_level = df_protein_level.groupby(['condition_1', 'condition_2','level'], group_keys=False).apply(mtcorrect).reset_index()
        df_edge_level = df_edge_level.groupby(['condition_1', 'condition_2','level'], group_keys=False).apply(mtcorrect).reset_index()
        df_edge = df_edge_level.sort_values('pvalue').groupby(['condition_1','condition_2','bait_id','prey_id']).head(1).reset_index()
        df_node_level = df_node_level.groupby(['condition_1', 'condition_2','level'], group_keys=False).apply(mtcorrect).reset_index()

        df_node_level_filtered = df_node_level[df_node_level['abs_log2fx'] > self.min_abs_log2fx]
        df_node = df_node_level_filtered.sort_values('pvalue_adjusted').groupby(['condition_1','condition_2','bait_id'], group_keys=False).head(1).reset_index()

        click.echo("Info: Total dysregulated proteins detected:")
        click.echo("%s (at FDR < 0.01)" % (df_node[df_node['pvalue_adjusted'] < 0.01][['bait_id']].drop_duplicates().shape[0]))
        click.echo("%s (at FDR < 0.05)" % (df_node[df_node['pvalue_adjusted'] < 0.05][['bait_id']].drop_duplicates().shape[0]))
        click.echo("%s (at FDR < 0.1)" % (df_node[df_node['pvalue_adjusted'] < 0.1][['bait_id']].drop_duplicates().shape[0]))

        for level in df_node_level_filtered['level'].unique():
            click.echo("Info: Dysregulated (%s-mode) proteins detected:" % (level))
            click.echo("%s (at FDR < 0.01)" % (df_node_level_filtered[(df_node_level_filtered['level'] == level) & (df_node_level_filtered['pvalue_adjusted'] < 0.01)][['bait_id']].drop_duplicates().shape[0]))
            click.echo("%s (at FDR < 0.05)" % (df_node_level_filtered[(df_node_level_filtered['level'] == level) & (df_node_level_filtered['pvalue_adjusted'] < 0.05)][['bait_id']].drop_duplicates().shape[0]))
            click.echo("%s (at FDR < 0.1)" % (df_node_level_filtered[(df_node_level_filtered['level'] == level) & (df_node_level_filtered['pvalue_adjusted'] < 0.1)][['bait_id']].drop_duplicates().shape[0]))

        return df_edge_level[['condition_1','condition_2','level','bait_id','prey_id','log2fx','abs_log2fx','interactor_ratio','pvalue','pvalue_adjusted']+[c for c in df_edge_level.columns if c.startswith("viper_")]], df_edge[['condition_1','condition_2','level','bait_id','prey_id','log2fx','abs_log2fx','interactor_ratio','pvalue','pvalue_adjusted']], df_node_level[['condition_1','condition_2','level','bait_id','log2fx','abs_log2fx','interactor_ratio','num_interactors','pvalue','pvalue_adjusted']], df_node[['condition_1','condition_2','level','bait_id','log2fx','abs_log2fx','interactor_ratio','num_interactors','pvalue','pvalue_adjusted']], df_protein_level[['condition_1','condition_2','level','bait_id','log2fx','abs_log2fx','interactor_ratio','pvalue','pvalue_adjusted'] + [c for c in df_protein_level.columns if c.startswith("viper_")]]


import pandas as pd
import numpy as np
import click
import sqlite3
import os
import sys

try:
    import matplotlib
    matplotlib.use('Agg')
    from matplotlib.backends.backend_pdf import PdfPages
    from matplotlib.lines import Line2D
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

def check_sqlite_table(con, table):
    table_present = False
    c = con.cursor()
    c.execute('SELECT count(name) FROM sqlite_master WHERE type="table" AND name="%s"' % table)
    if c.fetchone()[0] == 1:
        table_present = True
    else:
        table_present = False
    c.fetchall()

    return(table_present)

class plot_features:

    def __init__(self, infile, level, id, max_qvalue, min_abs_log2fx, mode, combined, peptide_rank):
        self.infile = infile
        self.level = level
        self.id = id
        self.max_qvalue = max_qvalue
        self.min_abs_log2fx = min_abs_log2fx
        self.mode = mode
        self.combined = combined
        self.peptide_rank = peptide_rank

        # Set global SEC boundaries
        self.sec_min, self.sec_max = self.read_sec_boundaries()

        # Read peptide and feature data
        self.feature_data = self.read_features()
        self.peptide_data = self.read_peptides()
        self.protein_data = self.read_proteins()

        # Read meta data if available
        self.interactions_dmeta = self.read_interactions_dmeta()
        self.interactions_qmeta = self.read_interactions_qmeta()

        self.monomer_qmeta = self.read_monomer_qmeta()

        if self.level == 'interaction' and self.id is not None:
            self.plot_interaction(self.id)
        elif self.level == 'bait' and self.id is not None:
            self.plot_bait(self.id, 0)
        elif self.level == 'interaction' and self.id is None and self.max_qvalue is not None:
            interaction_ids, result_ids, decoys = self.read_interactions()
            for interaction_id, result_id, decoy in zip(interaction_ids, result_ids, decoys):
                self.plot_interaction(interaction_id, result_id, decoy)
        elif self.level == 'bait' and self.id is None and self.max_qvalue is not None:
            bait_ids, result_ids = self.read_baits()
            for bait_id, result_id in zip(bait_ids, result_ids):
                self.plot_bait(bait_id, result_id)

    def plot_interaction(self, interaction_id, result_id=000000, decoy=False):
        feature_data = self.feature_data
        protein_data = self.protein_data
        protein_data['picked'] = True
        peptide_data = self.peptide_data

        bait_id, prey_id = interaction_id.split("_")

        feature_data = feature_data[feature_data['interaction_id'] == interaction_id]
        proteins = pd.DataFrame({"protein_id": [bait_id, prey_id]}).drop_duplicates()
        peptide_data = pd.merge(peptide_data, proteins, how='inner', on='protein_id')

        peptide_data = pd.merge(peptide_data, protein_data, on=['condition_id', 'replicate_id', 'protein_id', 'sec_id'], how='left')
        peptide_data.loc[peptide_data['picked'].isnull(), 'picked'] = False

        if decoy:
            out = os.path.splitext(os.path.basename(self.infile))[0]+"_"+str(result_id).zfill(6)+"_DECOY_"+interaction_id+".pdf"
        else:
            out = os.path.splitext(os.path.basename(self.infile))[0]+"_"+str(result_id).zfill(6)+"_"+interaction_id+".pdf"

        with PdfPages(out) as pdf:
            f = self.generate_plot(peptide_data, feature_data, bait_id, prey_id)
            pdf.savefig()
            plt.close()

    def plot_bait(self, bait_id, result_id):
        feature_data = self.feature_data
        protein_data = self.protein_data
        protein_data['picked'] = True
        peptide_data = self.peptide_data

        peptide_data = pd.merge(peptide_data, protein_data, on=['condition_id', 'replicate_id', 'protein_id', 'sec_id'], how='left')
        peptide_data.loc[peptide_data['picked'].isnull(), 'picked'] = False

        interaction_data = self.interactions_dmeta[['bait_id','prey_id','interaction_id']].drop_duplicates()
        interaction_data = interaction_data[(interaction_data['bait_id'] == bait_id) | (interaction_data['prey_id'] == bait_id)]

        # Add monomer
        interaction_data = pd.concat([pd.DataFrame({'bait_id': [bait_id], 'prey_id': [bait_id], 'interaction_id': [bait_id + "_" + bait_id]}), interaction_data], sort=False)

        out = os.path.splitext(os.path.basename(self.infile))[0]+"_"+str(result_id).zfill(6)+"_"+bait_id+".pdf"

        with PdfPages(out) as pdf:
            for idx, interaction in interaction_data.iterrows():
                feature_data_int = feature_data[feature_data['interaction_id'] == interaction['interaction_id']]
                if feature_data_int.shape[0] > 0:
                    proteins_int = pd.DataFrame({"protein_id": pd.concat([feature_data_int['bait_id'], feature_data_int['prey_id']])}).drop_duplicates()
                    peptide_data_int = pd.merge(peptide_data, proteins_int, how='inner', on='protein_id')

                    f = self.generate_plot(peptide_data_int, feature_data_int, feature_data_int['bait_id'].values[0], feature_data_int['prey_id'].values[0])
                    pdf.savefig()
                    plt.close()
                elif interaction['interaction_id'] == (bait_id + "_" + bait_id):
                    f = self.generate_plot(peptide_data[peptide_data['protein_id'] == bait_id], feature_data_int, bait_id, bait_id)
                    pdf.savefig()
                    plt.close()

    def read_sec_boundaries(self):
        con = sqlite3.connect(self.infile)

        df = pd.read_sql('SELECT min(sec_id) AS min_sec_id, max(sec_id) AS max_sec_id FROM SEC;', con)

        con.close()

        return df['min_sec_id'].values[0], df['max_sec_id'].values[0]

    def read_features(self):
        con = sqlite3.connect(self.infile)

        df = pd.read_sql('SELECT *, condition_id || "_" || replicate_id AS tag, bait_id || "_" || prey_id AS interaction_id FROM FEATURE_SCORED;', con)

        con.close()

        return df

    def read_proteins(self):
        con = sqlite3.connect(self.infile)

        df = pd.read_sql('SELECT * FROM PROTEIN_PEAKS;', con)

        con.close()

        return df

    def read_peptides(self):
        con = sqlite3.connect(self.infile)

        df = pd.read_sql('SELECT SEC.condition_id || "_" || SEC.replicate_id AS tag, SEC.condition_id, SEC.replicate_id, SEC.sec_id, QUANTIFICATION.protein_id, QUANTIFICATION.peptide_id, peptide_intensity, MONOMER.sec_id AS monomer_sec_id FROM QUANTIFICATION INNER JOIN PROTEIN_META ON QUANTIFICATION.protein_id = PROTEIN_META.protein_id INNER JOIN PEPTIDE_META ON QUANTIFICATION.peptide_id = PEPTIDE_META.peptide_id INNER JOIN SEC ON QUANTIFICATION.RUN_ID = SEC.RUN_ID INNER JOIN MONOMER ON QUANTIFICATION.protein_id = MONOMER.protein_id and SEC.condition_id = MONOMER.condition_id AND SEC.replicate_id = MONOMER.replicate_id WHERE peptide_rank <= %s;' % (self.peptide_rank), con)

        con.close()

        return df

    def read_interactions(self):
        con = sqlite3.connect(self.infile)

        if self.combined:
            table = 'EDGE'
        else:
            table = 'EDGE_LEVEL'

        if check_sqlite_table(con, 'EDGE') and self.mode == 'quantitative':
            df = pd.read_sql('SELECT DISTINCT bait_id || "_" || prey_id AS interaction_id, 0 as decoy FROM %s WHERE pvalue_adjusted < %s AND abs_log2fx > %s ORDER BY pvalue ASC;' % (table, self.max_qvalue, self.min_abs_log2fx), con)
        elif self.mode == 'detection':
            df = pd.read_sql('SELECT DISTINCT bait_id || "_" || prey_id AS interaction_id, decoy FROM FEATURE_SCORED_COMBINED WHERE qvalue < %s GROUP BY bait_id, prey_id ORDER BY qvalue ASC;' % (self.max_qvalue), con)
        else:
            sys.exit("Error: Mode for interaction plotting not supported.")

        con.close()

        return df['interaction_id'].values, df.index+1, df['decoy'].values

    def read_interactions_dmeta(self):
        con = sqlite3.connect(self.infile)

        if check_sqlite_table(con, 'COMPLEX_QM') and (self.mode == 'quantitative'):
            df = pd.read_sql('SELECT FEATURE_SCORED_COMBINED.bait_id AS bait_id, FEATURE_SCORED_COMBINED.prey_id AS prey_id, FEATURE_SCORED_COMBINED.bait_id || "_" || FEATURE_SCORED_COMBINED.prey_id AS interaction_id, BAIT_META.protein_name AS bait_name, PREY_META.protein_name AS prey_name, min(FEATURE_SCORED_COMBINED.pvalue) AS pvalue, min(FEATURE_SCORED_COMBINED.qvalue) AS qvalue FROM FEATURE_SCORED_COMBINED INNER JOIN (SELECT * FROM PROTEIN) AS BAIT_META ON FEATURE_SCORED_COMBINED.bait_id = BAIT_META.protein_id INNER JOIN (SELECT * FROM PROTEIN) AS PREY_META ON FEATURE_SCORED_COMBINED.prey_id = PREY_META.protein_id INNER JOIN (SELECT DISTINCT bait_id, prey_id FROM COMPLEX_QM) AS COMPLEX_QM ON FEATURE_SCORED_COMBINED.bait_id = COMPLEX_QM.bait_id AND FEATURE_SCORED_COMBINED.prey_id = COMPLEX_QM.prey_id GROUP BY FEATURE_SCORED_COMBINED.bait_id, FEATURE_SCORED_COMBINED.prey_id;', con)
        elif self.mode == 'detection':
            df = pd.read_sql('SELECT FEATURE_SCORED_COMBINED.bait_id AS bait_id, FEATURE_SCORED_COMBINED.prey_id AS prey_id, FEATURE_SCORED_COMBINED.bait_id || "_" || FEATURE_SCORED_COMBINED.prey_id AS interaction_id, BAIT_META.protein_name AS bait_name, PREY_META.protein_name AS prey_name, min(FEATURE_SCORED_COMBINED.pvalue) AS pvalue, min(FEATURE_SCORED_COMBINED.qvalue) AS qvalue FROM FEATURE_SCORED_COMBINED INNER JOIN (SELECT * FROM PROTEIN) AS BAIT_META ON FEATURE_SCORED_COMBINED.bait_id = BAIT_META.protein_id INNER JOIN (SELECT * FROM PROTEIN) AS PREY_META ON FEATURE_SCORED_COMBINED.prey_id = PREY_META.protein_id GROUP BY FEATURE_SCORED_COMBINED.bait_id, FEATURE_SCORED_COMBINED.prey_id;', con)
        else:
            df = None

        con.close()

        return df

    def read_interactions_qmeta(self):
        con = sqlite3.connect(self.infile)

        df = None

        if check_sqlite_table(con, 'EDGE'):
            df = pd.read_sql('SELECT condition_1, condition_2, bait_id, prey_id, pvalue, pvalue_adjusted, level, bait_id || "_" || prey_id AS interaction_id FROM EDGE_LEVEL;', con)
        else:
            df = None

        con.close()

        return df

    def read_monomer_qmeta(self):
        con = sqlite3.connect(self.infile)

        df = None

        if check_sqlite_table(con, 'NODE'):
            df = pd.read_sql('SELECT condition_1, condition_2, bait_id, pvalue, pvalue_adjusted, level FROM NODE_LEVEL;', con)
        else:
            df = None

        con.close()

        return df

    def read_baits(self):
        con = sqlite3.connect(self.infile)

        if not check_sqlite_table(con, 'NODE'):
            sys.exit("Error: Your experimental design is not supported. At least two conditions are necessary for differential analysis. Switch 'level' to 'interaction' for visualization.")

        if self.combined:
            table = 'NODE'
        else:
            table = 'NODE_LEVEL'

        if self.mode == 'quantitative':
            df = pd.read_sql('SELECT DISTINCT bait_id, min(pvalue) as pvalue FROM %s WHERE pvalue_adjusted < %s AND abs_log2fx > %s GROUP BY bait_id;' % (table, self.max_qvalue, self.min_abs_log2fx), con)

        con.close()

        df = df.sort_values(by=['pvalue']).reset_index()

        return df['bait_id'].values, df.index+1

    def generate_plot(self, peptide_data, feature_data, bait_id, prey_id):
        interaction_id = bait_id + "_" + prey_id

        tags = peptide_data.sort_values(by=['tag'])['tag'].drop_duplicates().values.tolist()

        f = plt.figure(figsize=(12,(len(tags)+1)*2.5))

        # Axes that share the x- and y-axes
        ax = f.add_subplot(len(tags)+1, 1, 1)
        axarr = [ax] + [f.add_subplot(len(tags)+1, 1, i, sharex=ax, sharey=ax) for i in range(2, len(tags)+1)]
        # The bottom independent subplot
        axarr.append(f.add_subplot(len(tags)+1, 1, len(tags)+1))

        # plot detection metadata
        if bait_id is not prey_id and self.interactions_dmeta is not None:
            dmeta = self.interactions_dmeta
            if dmeta[dmeta['interaction_id'] == interaction_id].shape[0] > 0:
                dmeta = dmeta[dmeta['interaction_id'] == interaction_id][['bait_name','prey_name','pvalue','qvalue']]
                titletext = str(interaction_id) + "\n" + str(dmeta['bait_name'].values[0]) + " vs "  + str(dmeta['prey_name'].values[0]) + "\n" + "p-value: "  + str(np.round(dmeta['pvalue'].values[0], 3)) + " q-value: "  + str(np.round(dmeta['qvalue'].values[0], 3))
            else:
                titletext = str(interaction_id)
        else:
            titletext = str(interaction_id)
        f.suptitle(titletext)

        ymin = 0 #peptide_data['peptide_intensity'].min()
        ymax = peptide_data['peptide_intensity'].max() * 1.2

        # plot interactions
        for tag in tags:
            axarr[tags.index(tag)].set_xlim(self.sec_min, self.sec_max)
            axarr[tags.index(tag)].set_ylim(ymin, ymax)
            proteins = peptide_data['protein_id'].drop_duplicates().values
            for protein in proteins:
                if protein == bait_id:
                    background_color = 'rosybrown'
                    protein_color = 'red'
                else:
                    background_color = 'grey'
                    protein_color = 'black'
                # plot monomer threshold
                axarr[tags.index(tag)].axvline(x=peptide_data[peptide_data['protein_id'] == protein]['monomer_sec_id'].mean(), color=protein_color, alpha=0.5)

                # plot peptide chromatograms
                peptides = peptide_data[peptide_data['protein_id'] == protein]['peptide_id'].drop_duplicates().values
                for peptide in peptides:
                    points = peptide_data[(peptide_data['peptide_id'] == peptide) & (peptide_data['tag'] == tag)].sort_values(by=['sec_id'])

                    if points.shape[0] > 0:
                        # background
                        axarr[tags.index(tag)].plot(points['sec_id'], points['peptide_intensity'], color=background_color)
                        # # picked proteins
                        axarr[tags.index(tag)].plot(points[points['picked']]['sec_id'], points[points['picked']]['peptide_intensity'], color=protein_color)

            # plot legend and subtitle
            if bait_id is not prey_id:
                axarr[tags.index(tag)].legend([Line2D([0], [0], color='red'), Line2D([0], [0], color='black')], [bait_id, prey_id], loc = 1)
            axarr[tags.index(tag)].set_title(tag, loc = 'center', pad = -15)

            # plot feature information if present
            feature = feature_data[(feature_data['bait_id'] == bait_id) & (feature_data['prey_id'] == prey_id) & (feature_data['tag'] == tag)]

            if feature.shape[0] > 0:
                feature_string = "p-value: %s\nq-value: %s\npep: %s" % (np.round(feature['pvalue'].mean(),3),np.round(feature['qvalue'].mean(),3),np.round(feature['pep'].mean(),3))
                axarr[tags.index(tag)].text(0.01, 0.95, feature_string, transform=axarr[tags.index(tag)].transAxes, fontsize=10, verticalalignment='top', bbox=dict(boxstyle='square', facecolor='wheat', alpha=0.5))

        # plot quantitative metadata
        if bait_id == prey_id:
            mmeta = self.monomer_qmeta
            mmeta = mmeta[mmeta['bait_id'] == bait_id][['level','condition_1','condition_2','pvalue','pvalue_adjusted']].sort_values(by='pvalue')
            if mmeta.shape[0] > 0:
                axarr[len(tags)].table(cellText=mmeta.values, colLabels=mmeta.columns, loc='center')
        elif self.interactions_qmeta is not None:
            qmeta = self.interactions_qmeta
            qmeta = qmeta[qmeta['interaction_id'] == interaction_id][['level','condition_1','condition_2','pvalue','pvalue_adjusted']].sort_values(by='pvalue')
            if qmeta.shape[0] > 0:
                axarr[len(tags)].table(cellText=qmeta.values, colLabels=qmeta.columns, loc='center')
        axarr[len(tags)].axis('off')

        return f





"""
Created on Sun Jun 21 12:22:12 2015

@author:  William Poole: wpoole@caltech.edu
"""

import numpy as np
from statsmodels.distributions.empirical_distribution import ECDF #empirical cumulative distribution function
from scipy.special import chdtrc as chi2_cdf
from scipy.stats import pearsonr


#Input: An m x n data matrix with each of m rows representing a variable and each of n columns representing a sample. Should be of type numpy.array
#       A vector of m P-values to combine. May be a list or of type numpy.array.
#Output: A combined P-value using the Empirical Brown's Method (EBM).
#        If extra_info == True: also returns the p-value from Fisher's method, the scale factor c, and the new degrees of freedom from Brown's Method
def EmpiricalBrownsMethod(data_matrix, p_values, extra_info = False):
    covar_matrix = CalculateCovariances(data_matrix)
    return CombinePValues(covar_matrix, p_values, extra_info)


#Input: raw data vector (of one variable) with no missing samples. May be a list or an array.
#Output Transforemd data vector w.
def TransformData(data_vector):
    m = np.mean(data_vector)
    sd = np.std(data_vector)
    s = [(d-m)/sd for d in data_vector]
    W = lambda x: -2*np.log(ECDF(s)(x))
    return np.array([W(x) for x in s])

#Input: An m x n data matrix with each of m rows representing a variable and each of n columns representing a sample. Should be of type numpy.array.
#       Note: Method does not deal with missing values within the data.
#Output: An m x m matrix of pairwise covariances between transformed raw data vectors
def CalculateCovariances(data_matrix):
    transformed_data_matrix = np.array([TransformData(f) for f in data_matrix])
    covar_matrix = np.cov(transformed_data_matrix)

    return covar_matrix
    
#Input: A m x m numpy array of covariances between transformed data vectors and a vector of m p-values to combine.
#Output: A combined P-value. 
#        If extra_info == True: also returns the p-value from Fisher's method, the scale factor c, and the new degrees of freedom from Brown's Method
def CombinePValues(covar_matrix, p_values, extra_info = False):
    m = int(covar_matrix.shape[0])
    #print "m", m
    df_fisher = 2.0*m
    Expected = 2.0*m
    cov_sum = 0
    for i in range(m):
        for j in range(i+1, m):
            cov_sum += covar_matrix[i, j]
    
    #print "cov sum", cov_sum
    Var = 4.0*m+2*cov_sum
    c = Var/(2.0*Expected)
    df_brown = 2.0*Expected**2/Var
    if df_brown > df_fisher:
        df_brown = df_fisher
        c = 1.0

    x = 2.0*sum([-np.log(p) for p in p_values])
    #print "x", x
    p_brown = chi2_cdf(df_brown, 1.0*x/c)
    p_fisher = chi2_cdf(df_fisher, 1.0*x)
    
    if extra_info:
        return p_brown, p_fisher, c, df_brown
    else:
        return p_brown

#Input: An m x n data matrix with each of m rows representing a variable and each of n columns representing a sample. Should be of type numpy.array
#       A vector of m P-values to combine. May be a list or of type numpy.array.
#Output: A combined P-value using Kost's Method.
#        If extra_info == True: also returns the p-value from Fisher's method, the scale factor c, and the new degrees of freedom from Brown's Method
def KostsMethod(data_matrix, p_values, extra_info = False):
    covar_matrix = CalculateKostCovariance(data_matrix)
    return CombinePValues(covar_matrix, p_values, extra_info = extra_info)
    
#Input correlation between two n x n data vectors.
#Output: Kost's approximation of the covariance between the -log cumulative distributions. This is calculated with a cubic polynomial fit.
def KostPolyFit(cor):
    a1, a2, a3 = 3.263, .710, .027 #Kost cubic coeficients
    return a1*cor+a2*cor**2+a3*cor**3
    
#Input: An m x n data matrix with each of m rows representing a variable and each of n columns representing a sample. Should be of type numpy.array.
#       Note: Method does not deal with missing values within the data.
#Output: An m x m matrix of pairwise covariances between the data vectors calculated using Kost's polynomial fit and numpy's pearson correlation function.
def CalculateKostCovariance(data_matrix):
    m = data_matrix.shape[0]
    covar_matrix = np.zeros((m, m))
    for i in range(m):
        for j in range(i+1, m):
            cor, p_val = pearsonr(data_matrix[i, :], data_matrix[j, :])
            covar = KostPolyFit(cor)
            covar_matrix[i, j] = covar
            covar_matrix[j, i] = covar
    return covar_matrix
    
    
    
    



        
    


import pandas as pd
import numpy as np
import scipy as sp
import click
import sqlite3
import os
import sys

import multiprocessing
from functools import partial
from tqdm import tqdm

from scipy.signal import find_peaks, peak_widths
from minepy import cstats

# np.seterr(divide='ignore', invalid='ignore')
# np.seterr(all='raise')

# Find Monomer Threshold
class monomer:
    def __init__(self, outfile, monomer_threshold_factor):
        self.outfile = outfile
        self.monomer_threshold_factor = monomer_threshold_factor
        self.df = self.protein_thresholds()

    def protein_thresholds(self):
        def get_sec_ids(protein_mw, sec_meta):
            def condition_replicate_sec_ids(protein_mw, run_sec_meta):
                return pd.Series({'sec_id': (run_sec_meta['sec_mw']-self.monomer_threshold_factor*protein_mw).abs().argsort().iloc[:1].values[0]})

            return sec_meta.groupby(['condition_id','replicate_id']).apply(lambda x: condition_replicate_sec_ids(protein_mw, x[['sec_id','sec_mw']])).reset_index()

        con = sqlite3.connect(self.outfile)
        protein_mw = pd.read_sql('SELECT protein_id, protein_mw FROM PROTEIN;', con)
        sec_meta = pd.read_sql('SELECT DISTINCT condition_id, replicate_id, sec_id, sec_mw FROM SEC;', con)
        con.close()

        # Compute expected SEC fraction
        protein_sec_thresholds = protein_mw.groupby(['protein_id']).apply(lambda x: get_sec_ids(x['protein_mw'].mean(), sec_meta=sec_meta)).reset_index(level=['protein_id'])

        return protein_sec_thresholds[['condition_id','replicate_id','protein_id','sec_id']]

def score_chunk(queries, qm, run):
    scores = []
    for query_ix, query in queries.iterrows():
        bait = qm.xs(query['bait_id'], level='protein_id')
        bait_monomer_sec_id = bait.iloc[0].name[1]

        prey = qm.xs(query['prey_id'], level='protein_id')
        prey_monomer_sec_id = prey.iloc[0].name[1]

        score = score_interaction(bait.values.copy(), prey.values.copy(), bait_monomer_sec_id, prey_monomer_sec_id)
        if score is not None:
            score['condition_id'] = run['condition_id']
            score['replicate_id'] = run['replicate_id']
            score = {**score, **query}
            scores.append(score)
    return(scores)

def score_interaction(bait, prey, bait_monomer_sec_id, prey_monomer_sec_id):
    def longest_intersection(arr):
        # Compute longest continuous stretch
        n = len(arr)
        s = set()
        ans=0
        for ele in arr:
            s.add(ele)
        for i in range(n):
            if (arr[i]-1) not in s:
                j=arr[i]
                while(j in s):
                    j+=1
                ans=max(ans, j-arr[i])
        return ans

    def normalized_xcorr(a, b):
        # Normalize matrices
        a = (a - np.mean(a, axis=1, keepdims=True)) / (np.std(a, axis=1, keepdims=True))
        b = (b - np.mean(b, axis=1, keepdims=True)) / (np.std(b, axis=1, keepdims=True))

        nxcorr = [] # normalized cross-correlation
        lxcorr = [] # cross-correlation lag

        if np.array_equal(a,b):
            # Compare all rows of a against all rows of a, including itself (auto-correlation)
            for i in range(0, len(a)):
                for j in range(i, len(a)):
                    nxcorr.append(np.correlate(a[i], a[j], 'valid')[0] / len(a[i])) # Normalize by length
                    lxcorr.append(np.argmax(np.correlate(a[i], a[j], 'same'))) # Peak
        else:
            # Compare all rows of a against all rows of b
            for i in range(0, len(a)):
                for j in range(0, len(b)):
                    nxcorr.append(np.correlate(a[i], b[j], 'valid')[0] / len(a[i])) # Normalize by length
                    lxcorr.append(np.argmax(np.correlate(a[i], b[j], 'same'))) # Peak

        return np.array(nxcorr), np.array(lxcorr)        

    def sec_xcorr(bm, pm):
        # Compute SEC xcorr scores
        bnx, blx = normalized_xcorr(bm, bm)
        pnx, plx = normalized_xcorr(pm, pm)
        bpnx, bplx = normalized_xcorr(bm, pm)

        xcorr_shape = np.mean(bpnx)
        xcorr_apex = np.mean(bplx)
        xcorr_shift = max([abs(xcorr_apex - np.mean(blx)), abs(xcorr_apex - np.mean(plx))])

        return xcorr_shape, xcorr_shift, xcorr_apex

    def mass_similarity(bm, pm):
        # Sum bait and prey peptides
        bpabundance = np.sum(bm, axis=1, keepdims=True).mean()
        ppabundance = np.sum(pm, axis=1, keepdims=True).mean()

        # Compute abundance ratio of bait and prey protein
        abundance_ratio = bpabundance / ppabundance
        if abundance_ratio > 1:
            abundance_ratio = 1 / abundance_ratio

        return abundance_ratio


   # Compute bait and prey overlap
    overlap = (np.nansum(bait, axis=0) > 0) | (np.nansum(prey, axis=0) > 0)
    total_overlap = np.count_nonzero(overlap)

    # Compute bait and prey intersection
    intersection = (np.nansum(bait, axis=0) > 0) & (np.nansum(prey, axis=0) > 0)
    total_intersection = np.count_nonzero(intersection)
    if total_intersection > 0:
        longest_intersection = longest_intersection(intersection.nonzero()[0])

        # Require at least three consecutive overlapping data points
        if longest_intersection > 2:
            # Prepare total bait and prey profiles & Replace nan with 0
            total_bait = np.nan_to_num(bait)
            total_prey = np.nan_to_num(prey)

            # Remove non-overlapping segments
            bait[:,~intersection] = np.nan
            prey[:,~intersection] = np.nan

            # Remove completely empty peptides
            bait = bait[(np.nansum(bait,axis=1) > 0),:]
            prey = prey[(np.nansum(prey,axis=1) > 0),:]

            # Replace nan with 0
            bait = np.nan_to_num(bait)
            prey = np.nan_to_num(prey)

            # Require at least one remaining peptide for bait and prey
            if (bait.shape[0] > 0) and (prey.shape[0] > 0):
                # Compute cross-correlation scores
                xcorr_shape, xcorr_shift, xcorr_apex = sec_xcorr(bait, prey)

                # Compute MIC/TIC scores
                mic_stat, tic_stat = cstats(bait[:,intersection], prey[:,intersection], est="mic_e")
                mic = mic_stat.mean(axis=0).mean() # Axis 0: summary for prey peptides / Axis 1: summary for bait peptides
                tic = tic_stat.mean(axis=0).mean() # Axis 0: summary for prey peptides / Axis 1: summary for bait peptides

                # Compute mass similarity score
                abundance_ratio = mass_similarity(bait, prey)

                # Compute total mass similarity score
                total_abundance_ratio = mass_similarity(total_bait, total_prey)

                # Compute relative intersection score
                relative_overlap = total_intersection / total_overlap

                # Compute delta monomer score
                delta_monomer = np.abs(bait_monomer_sec_id - prey_monomer_sec_id)

                # Compute apex monomer score
                apex_monomer = np.min(np.array(bait_monomer_sec_id - xcorr_apex, prey_monomer_sec_id - xcorr_apex))

                return({'var_xcorr_shape': xcorr_shape, 'var_xcorr_shift': xcorr_shift, 'var_abundance_ratio': abundance_ratio, 'var_total_abundance_ratio': total_abundance_ratio, 'var_mic': mic, 'var_tic': tic, 'var_sec_overlap': relative_overlap, 'var_sec_intersection': longest_intersection, 'var_delta_monomer': delta_monomer, 'var_apex_monomer': apex_monomer})

# Scoring
class scoring:
    def __init__(self, outfile, chunck_size, threads, minimum_peptides, maximum_peptides, peakpicking):
        self.outfile = outfile
        self.chunck_size = chunck_size
        self.threads = threads
        self.minimum_peptides = minimum_peptides
        self.maximum_peptides = maximum_peptides
        self.peakpicking = peakpicking

        self.sec_boundaries = self.read_sec_boundaries()

        click.echo("Info: Read peptide chromatograms.")
        chromatograms = self.read_chromatograms()
        click.echo("Info: Filter peptide chromatograms.")
        self.chromatograms = self.filter_peptides(chromatograms)
        self.store_filtered()
        click.echo("Info: Read queries and SEC boundaries.")
        self.queries = self.read_queries()

        click.echo("Info: Score PPI.")
        self.compare()

    def read_chromatograms(self):
        # Read data
        con = sqlite3.connect(self.outfile)
        df = pd.read_sql('SELECT SEC.condition_id, SEC.replicate_id, SEC.sec_id, QUANTIFICATION.protein_id, QUANTIFICATION.peptide_id, peptide_intensity, MONOMER.sec_id AS monomer_sec_id FROM QUANTIFICATION INNER JOIN PROTEIN_META ON QUANTIFICATION.protein_id = PROTEIN_META.protein_id INNER JOIN PEPTIDE_META ON QUANTIFICATION.peptide_id = PEPTIDE_META.peptide_id INNER JOIN SEC ON QUANTIFICATION.RUN_ID = SEC.RUN_ID INNER JOIN MONOMER ON QUANTIFICATION.protein_id = MONOMER.protein_id and SEC.condition_id = MONOMER.condition_id AND SEC.replicate_id = MONOMER.replicate_id WHERE peptide_count >= %s AND peptide_rank <= %s;' % (self.minimum_peptides, self.maximum_peptides), con)

        con.close()

        return df

    def filter_peptides(self, df):
        def peptide_detrend_zero(x):
            peptide_mean = np.mean(np.append(x['peptide_intensity'], np.zeros(len(self.sec_boundaries['sec_id'].unique())-x.shape[0])))
            return x[x['peptide_intensity'] > peptide_mean][['sec_id','peptide_intensity','monomer_sec_id']]

        def peptide_detrend_drop(x):
            peptide_mean = np.mean(x['peptide_intensity'])
            return x[x['peptide_intensity'] > peptide_mean][['sec_id','peptide_intensity','monomer_sec_id']]

        def protein_pick(x):
            xpep = x.groupby(['peptide_id','sec_id'])['peptide_intensity'].mean().reset_index()
            xprot = xpep.groupby(['sec_id'])['peptide_intensity'].mean().reset_index()

            xall = pd.merge(self.sec_boundaries, xprot[['sec_id','peptide_intensity']], on='sec_id', how='left').sort_values(['sec_id'])
            xall['peptide_intensity'] = np.nan_to_num(xall['peptide_intensity'].values) # Replace missing values with zeros

            peaks, _ = find_peaks(xall['peptide_intensity'], width=[3,])
            boundaries = peak_widths(xall['peptide_intensity'], peaks, rel_height=0.9)

            left_boundaries = np.floor(boundaries[2])
            right_boundaries = np.ceil(boundaries[3])

            sec_list = None
            for peak in list(zip(left_boundaries, right_boundaries)):
                if sec_list is None:
                    sec_list = np.arange(peak[0],peak[1]+1)
                else:
                    sec_list = np.append(sec_list, np.arange(peak[0],peak[1]+1))

            if len(x['replicate_id'].unique()) == 1:
                return x[x['sec_id'].isin(np.unique(sec_list))][['peptide_id','sec_id','peptide_intensity','monomer_sec_id']]
            else:
                return x[x['sec_id'].isin(np.unique(sec_list))][['replicate_id','peptide_id','sec_id','peptide_intensity','monomer_sec_id']]

        # Report statistics before filtering
        click.echo("Info: %s unique peptides before filtering." % len(df['peptide_id'].unique()))
        click.echo("Info: %s peptide chromatograms before filtering." % df[['condition_id','replicate_id','protein_id','peptide_id']].drop_duplicates().shape[0])
        click.echo("Info: %s data points before filtering." % df.shape[0])

        # Filter monomers
        df = df[df['sec_id'] <= df['monomer_sec_id']]

        if self.peakpicking == "detrend_zero":
            # Remove constant trends from peptides, average over all fractions
            df = df.groupby(['condition_id','replicate_id','protein_id','peptide_id']).apply(peptide_detrend_zero).reset_index(level=['condition_id','replicate_id','protein_id','peptide_id'])
        if self.peakpicking == "detrend_drop":
            # Remove constant trends from peptides, average over fractions with detections
            df = df.groupby(['condition_id','replicate_id','protein_id','peptide_id']).apply(peptide_detrend_drop).reset_index(level=['condition_id','replicate_id','protein_id','peptide_id'])
        elif self.peakpicking == "localmax_conditions":
            # Protein-level peakpicking
            df = df.groupby(['condition_id','protein_id']).apply(protein_pick).reset_index(level=['condition_id','protein_id'])
        elif self.peakpicking == "localmax_replicates":
            # Protein-level peakpicking
            df = df.groupby(['condition_id','replicate_id','protein_id']).apply(protein_pick).reset_index(level=['condition_id','replicate_id','protein_id'])

        # Report statistics after filtering
        click.echo("Info: %s unique peptides after filtering." % len(df['peptide_id'].unique()))
        click.echo("Info: %s peptide chromatograms after filtering." % df[['condition_id','replicate_id','protein_id','peptide_id']].drop_duplicates().shape[0])
        click.echo("Info: %s data points after filtering." % df.shape[0])

        return df

    def store_filtered(self):
        con = sqlite3.connect(self.outfile)
        self.chromatograms[['condition_id','replicate_id','protein_id','sec_id']].drop_duplicates().to_sql('PROTEIN_PEAKS', con, index=False, if_exists='replace')
        con.close()

    def read_queries(self):
        # Read data
        con = sqlite3.connect(self.outfile)
        df = pd.read_sql('SELECT * FROM QUERY;', con)
        con.close()

        return df

    def read_sec_boundaries(self):
        # Read data
        con = sqlite3.connect(self.outfile)
        df = pd.read_sql('SELECT min(sec_id) AS min_sec_id, max(sec_id) AS max_sec_id FROM SEC;', con)
        con.close()

        return pd.DataFrame({'sec_id': range(df['min_sec_id'].values[0], df['max_sec_id'].values[0]+1)})

    def split_chunks(self, dfm):
        def index_marks(nrows, chunck_size):
            return range(1 * chunck_size, (nrows // chunck_size + 1) * chunck_size, chunck_size)

        indices = index_marks(dfm.shape[0], self.chunck_size)
        return np.split(dfm, indices)

    def compare(self):
        # Obtain experimental design
        exp_design = self.chromatograms[['condition_id','replicate_id']].drop_duplicates()

        # Iterate over experimental design
        for exp_ix, run in exp_design.iterrows():
            chromatograms = self.chromatograms[(self.chromatograms['condition_id']==run['condition_id']) & (self.chromatograms['replicate_id']==run['replicate_id'])]
            qm = chromatograms.pivot_table(index=['protein_id','peptide_id','monomer_sec_id'], columns='sec_id', values='peptide_intensity')

            # Ensure that all queries are covered by chromatograms
            proteins = chromatograms['protein_id'].unique()
            queries = self.queries[self.queries['bait_id'].isin(proteins) & self.queries['prey_id'].isin(proteins)]

            # Split data into chunks for parallel processing
            queries_chunks = self.split_chunks(queries)
            click.echo("Info: Total number of queries for condition %s and replicate %s: %s. Split into %s chuncks." % (run['condition_id'], run['replicate_id'], queries.shape[0], len(queries_chunks)))
            
            # Initialize multiprocessing
            pool = multiprocessing.Pool(processes=self.threads)

            with tqdm(total=len(queries_chunks)) as pbar:
                for i, result in tqdm(enumerate(pool.imap_unordered(partial(score_chunk, qm=qm, run=run), queries_chunks))):

                    con = sqlite3.connect(self.outfile)
                    pd.DataFrame(result).to_sql('FEATURE', con, index=False, if_exists='append')
                    con.close()

                    pbar.update()



from pdb import set_trace
import pandas as pd
import numpy as np
import click
import sqlite3
import os
from os import path
import sys
from sqlite3 import connect
from tqdm import tqdm
from .plot import check_sqlite_table

class export_tables:
    def __init__(self, infile, level, id, max_qvalue, min_abs_log2fx, mode, combined, peptide_rank, extra=True):
        self.infile = infile
        self.level = level
        self.id = id
        self.max_qvalue = max_qvalue
        self.min_abs_log2fx = min_abs_log2fx
        self.mode = mode
        self.combined = combined
        self.peptide_rank = peptide_rank
        
        con = connect(infile)
        click.echo("Exporting basic data...")
        self.export_basic_data(con)
        
        if extra:
            click.echo("Exporting extra data...")
            # Set global SEC boundaries
            self.sec_min, self.sec_max = self.read_sec_boundaries(con)
            
            # Read peptide and feature data
            self.feature_data = self.read_features(con)
            self.peptide_data = self.read_peptides(con)
            self.protein_data = self.read_proteins(con)

            # Read meta data if available
            self.interactions_dmeta = self.read_interactions_dmeta(con)
            self.interactions_qmeta = self.read_interactions_qmeta(con)
            
            self.monomer_qmeta = self.read_monomer_qmeta(con)
            
            if self.level == 'interaction' and self.id is not None:
                self.export_interaction(self.id)
            elif self.level == 'bait' and self.id is not None:
                self.export_bait(self.id, 0)
            elif self.level == 'interaction' and self.id is None and self.max_qvalue is not None:
                interaction_ids, result_ids, decoys = self.read_interactions(con)
                click.echo(f"No interaction specified so exporting all {len(interaction_ids)} interactions")
                progress_bar = tqdm(zip(interaction_ids, result_ids, decoys), desc="Interactions", total=len(interaction_ids))
                
                for interaction_id, result_id, decoy in progress_bar:
                    self.export_interaction(interaction_id, result_id, decoy)
            elif self.level == 'bait' and self.id is None and self.max_qvalue is not None:
                bait_ids, result_ids = self.read_baits(con)
                click.echo(f"No bait specified so exporting all {len(bait_ids)} baits")
                progress_bar = tqdm(zip(bait_ids, result_ids), total=len(bait_ids), desc="Baits")
                
                for bait_id, result_id in progress_bar:
                    self.export_bait(bait_id, result_id)
                    
        con.close()
    
    def export_basic_data(self, con):
        outfile_interactions = path.splitext(self.infile)[0] + "_interactions.csv"
        outfile_network = path.splitext(self.infile)[0] + "_network.csv"
        outfile_nodes = path.splitext(self.infile)[0] + "_differential_nodes.csv"
        outfile_nodes_level = path.splitext(self.infile)[0] + "_differential_nodes_level.csv"
        outfile_edges = path.splitext(self.infile)[0] + "_differential_edges.csv"
        outfile_edges_level = path.splitext(self.infile)[0] + "_differential_edges_level.csv"
        outfile_proteins_level = path.splitext(self.infile)[0] + "_differential_proteins_level.csv"
        
        if check_sqlite_table(con, 'FEATURE_SCORED_COMBINED'):
            interaction_data = pd.read_sql('SELECT DISTINCT bait_id, prey_id FROM FEATURE_SCORED_COMBINED WHERE decoy == 0 and qvalue <= %s;' % self.max_qvalue , con)
            interaction_data.to_csv(outfile_interactions, index=False)
        if check_sqlite_table(con, 'FEATURE_SCORED_COMBINED') and check_sqlite_table(con, 'MONOMER_QM'):
            network_data = pd.read_sql('SELECT DISTINCT bait_id, prey_id FROM FEATURE_SCORED_COMBINED WHERE decoy == 0 and qvalue <= %s UNION SELECT DISTINCT bait_id, prey_id FROM MONOMER_QM;' % self.max_qvalue , con)
            network_data.to_csv(outfile_network, index=False)
        if check_sqlite_table(con, 'NODE'):
            node_data = pd.read_sql('SELECT * FROM NODE LEFT OUTER JOIN PROTEIN ON bait_id = protein_id;' , con)
            node_data.sort_values(by=['pvalue']).to_csv(outfile_nodes, index=False)
        if check_sqlite_table(con, 'NODE_LEVEL'):
            node_level_data = pd.read_sql('SELECT * FROM NODE_LEVEL LEFT OUTER JOIN PROTEIN ON bait_id = protein_id;' , con)
            node_level_data.sort_values(by=['pvalue']).to_csv(outfile_nodes_level, index=False)
        if check_sqlite_table(con, 'EDGE'):
            edge_data = pd.read_sql('SELECT * FROM EDGE;' , con)
            edge_data.sort_values(by=['pvalue']).to_csv(outfile_edges, index=False)
        if check_sqlite_table(con, 'EDGE_LEVEL'):
            edge_level_data = pd.read_sql('SELECT * FROM EDGE_LEVEL;' , con)
            edge_level_data.sort_values(by=['pvalue']).to_csv(outfile_edges_level, index=False)
        if check_sqlite_table(con, 'PROTEIN_LEVEL'):
            protein_level_data = pd.read_sql('SELECT * FROM PROTEIN_LEVEL LEFT OUTER JOIN PROTEIN ON bait_id = protein_id;' , con)
            protein_level_data.sort_values(by=['pvalue']).to_csv(outfile_proteins_level, index=False)
            
    
    def export_interaction(self, interaction_id, result_id=000000, decoy=False):
        feature_data = self.feature_data
        protein_data = self.protein_data
        protein_data['picked'] = True
        peptide_data = self.peptide_data

        bait_id, prey_id = interaction_id.split("_")

        feature_data = feature_data[feature_data['interaction_id'] == interaction_id]
        proteins = pd.DataFrame({"protein_id": [bait_id, prey_id]}).drop_duplicates()
        peptide_data = pd.merge(peptide_data, proteins, how='inner', on='protein_id')

        peptide_data = pd.merge(peptide_data, protein_data, on=['condition_id', 'replicate_id', 'protein_id', 'sec_id'], how='left')
        peptide_data.loc[peptide_data['picked'].isnull(), 'picked'] = False

        if decoy:
            peptide_out = os.path.splitext(os.path.basename(self.infile))[0]+"_DECOY_"+interaction_id+"_peptide"+".csv"
            feature_out = os.path.splitext(os.path.basename(self.infile))[0]+"_DECOY_"+interaction_id+"_feature"+".csv"
        else:
            peptide_out = os.path.splitext(os.path.basename(self.infile))[0]+"_"+interaction_id+"_peptide"+".csv"
            feature_out = os.path.splitext(os.path.basename(self.infile))[0]+"_"+interaction_id+"_feature"+".csv"
            
        peptide_data.to_csv(peptide_out, index=False)
        feature_data.to_csv(feature_out, index=False)
    
    def export_bait(self, bait_id, result_id):
        feature_data = self.feature_data
        protein_data = self.protein_data
        protein_data['picked'] = True
        peptide_data = self.peptide_data

        peptide_data = pd.merge(peptide_data, protein_data, on=['condition_id', 'replicate_id', 'protein_id', 'sec_id'], how='left')
        peptide_data.loc[peptide_data['picked'].isnull(), 'picked'] = False

        interaction_data = self.interactions_dmeta[['bait_id','prey_id','interaction_id']].drop_duplicates()
        interaction_data = interaction_data[(interaction_data['bait_id'] == bait_id) | (interaction_data['prey_id'] == bait_id)]

        # Add monomer
        interaction_data = pd.concat([pd.DataFrame({'bait_id': [bait_id], 'prey_id': [bait_id], 'interaction_id': [bait_id + "_" + bait_id]}), interaction_data], sort=False)
        peptide_dfs = []
        feature_dfs = []
        for idx, interaction in interaction_data.iterrows():
            feature_data_int = feature_data[feature_data['interaction_id'] == interaction['interaction_id']].copy()
            
            if feature_data_int.shape[0] > 0:
                proteins_int = pd.DataFrame({"protein_id": pd.concat([feature_data_int['bait_id'], feature_data_int['prey_id']])}).drop_duplicates()
                peptide_data_int = pd.merge(peptide_data, proteins_int, how='inner', on='protein_id') 
                peptide_data_int['picked'] = peptide_data_int['picked'].astype('bool')
                
                prey_id = feature_data_int['prey_id'].values[0]
                interaction_id = bait_id + "_" + prey_id
                peptide_data_int['interaction_id'] = interaction_id
                feature_data_int['interaction_id'] = interaction_id
                
                peptide_dfs.append(peptide_data_int)
                feature_dfs.append(feature_data_int)
                
            elif interaction['interaction_id'] == (bait_id + "_" + bait_id):
                peptide_data_int = peptide_data[peptide_data['protein_id'] == bait_id].copy()
                peptide_data_int['picked'] = peptide_data_int['picked'].astype('bool')
                
                interaction_id = bait_id + "_" + bait_id
                peptide_data_int['interaction_id'] = interaction_id
                feature_data_int['interaction_id'] = interaction_id
                
                peptide_dfs.append(peptide_data_int)
                feature_dfs.append(feature_data_int)
        
        peptide_df = pd.concat(peptide_dfs)
        feature_df = pd.concat(feature_dfs)
        peptide_out = os.path.splitext(os.path.basename(self.infile))[0]+"_"+bait_id+"_peptide"+".csv"
        feature_out = os.path.splitext(os.path.basename(self.infile))[0]+"_"+bait_id+"_feature"+".csv"                
        
        peptide_df.to_csv(peptide_out, index=False)
        feature_df.to_csv(feature_out, index=False)
    
    def read_sec_boundaries(self, con):
        df = pd.read_sql('SELECT min(sec_id) AS min_sec_id, max(sec_id) AS max_sec_id FROM SEC;', con)
        return df['min_sec_id'].values[0], df['max_sec_id'].values[0]

    def read_features(self, con):
        df = pd.read_sql(
            """
            SELECT *, 
            condition_id || "_" || replicate_id AS tag, 
            bait_id || "_" || prey_id AS interaction_id 
            FROM FEATURE_SCORED;
            """, 
            con
        )
        return df

    def read_proteins(self, con):
        df = pd.read_sql('SELECT * FROM PROTEIN_PEAKS;', con)
        return df

    def read_peptides(self, con):
        df = pd.read_sql(
            """
            SELECT 
                SEC.condition_id || "_" || SEC.replicate_id AS tag, 
                SEC.condition_id, SEC.replicate_id,
                SEC.sec_id, 
                QUANTIFICATION.protein_id, 
                QUANTIFICATION.peptide_id, 
                peptide_intensity, 
                MONOMER.sec_id AS monomer_sec_id 
            FROM QUANTIFICATION 
            INNER JOIN PROTEIN_META ON QUANTIFICATION.protein_id = PROTEIN_META.protein_id 
            INNER JOIN PEPTIDE_META ON QUANTIFICATION.peptide_id = PEPTIDE_META.peptide_id 
            INNER JOIN SEC ON QUANTIFICATION.RUN_ID = SEC.RUN_ID 
            INNER JOIN MONOMER ON QUANTIFICATION.protein_id = MONOMER.protein_id AND SEC.condition_id = MONOMER.condition_id AND SEC.replicate_id = MONOMER.replicate_id 
            WHERE peptide_rank <= %s;
            """ % (self.peptide_rank), 
            con
        )
        return df

    def read_interactions(self, con):
        if self.combined:
            table = 'EDGE'
        else:
            table = 'EDGE_LEVEL'

        if check_sqlite_table(con, 'EDGE') and self.mode == 'quantitative':
            df = pd.read_sql('SELECT DISTINCT bait_id || "_" || prey_id AS interaction_id, 0 as decoy FROM %s WHERE pvalue_adjusted < %s AND abs_log2fx > %s ORDER BY pvalue ASC;' % (table, self.max_qvalue, self.min_abs_log2fx), con)
        elif self.mode == 'detection':
            df = pd.read_sql('SELECT DISTINCT bait_id || "_" || prey_id AS interaction_id, decoy FROM FEATURE_SCORED_COMBINED WHERE qvalue < %s GROUP BY bait_id, prey_id ORDER BY qvalue ASC;' % (self.max_qvalue), con)
        else:
            sys.exit("Error: Mode for interaction plotting not supported.")

        return df['interaction_id'].values, df.index+1, df['decoy'].values

    def read_interactions_dmeta(self, con):
        if check_sqlite_table(con, 'COMPLEX_QM') and (self.mode == 'quantitative'):
            df = pd.read_sql('SELECT FEATURE_SCORED_COMBINED.bait_id AS bait_id, FEATURE_SCORED_COMBINED.prey_id AS prey_id, FEATURE_SCORED_COMBINED.bait_id || "_" || FEATURE_SCORED_COMBINED.prey_id AS interaction_id, BAIT_META.protein_name AS bait_name, PREY_META.protein_name AS prey_name, min(FEATURE_SCORED_COMBINED.pvalue) AS pvalue, min(FEATURE_SCORED_COMBINED.qvalue) AS qvalue FROM FEATURE_SCORED_COMBINED INNER JOIN (SELECT * FROM PROTEIN) AS BAIT_META ON FEATURE_SCORED_COMBINED.bait_id = BAIT_META.protein_id INNER JOIN (SELECT * FROM PROTEIN) AS PREY_META ON FEATURE_SCORED_COMBINED.prey_id = PREY_META.protein_id INNER JOIN (SELECT DISTINCT bait_id, prey_id FROM COMPLEX_QM) AS COMPLEX_QM ON FEATURE_SCORED_COMBINED.bait_id = COMPLEX_QM.bait_id AND FEATURE_SCORED_COMBINED.prey_id = COMPLEX_QM.prey_id GROUP BY FEATURE_SCORED_COMBINED.bait_id, FEATURE_SCORED_COMBINED.prey_id;', con)
        elif self.mode == 'detection':
            df = pd.read_sql('SELECT FEATURE_SCORED_COMBINED.bait_id AS bait_id, FEATURE_SCORED_COMBINED.prey_id AS prey_id, FEATURE_SCORED_COMBINED.bait_id || "_" || FEATURE_SCORED_COMBINED.prey_id AS interaction_id, BAIT_META.protein_name AS bait_name, PREY_META.protein_name AS prey_name, min(FEATURE_SCORED_COMBINED.pvalue) AS pvalue, min(FEATURE_SCORED_COMBINED.qvalue) AS qvalue FROM FEATURE_SCORED_COMBINED INNER JOIN (SELECT * FROM PROTEIN) AS BAIT_META ON FEATURE_SCORED_COMBINED.bait_id = BAIT_META.protein_id INNER JOIN (SELECT * FROM PROTEIN) AS PREY_META ON FEATURE_SCORED_COMBINED.prey_id = PREY_META.protein_id GROUP BY FEATURE_SCORED_COMBINED.bait_id, FEATURE_SCORED_COMBINED.prey_id;', con)
        else:
            df = None

        return df

    def read_interactions_qmeta(self, con):
        df = None

        if check_sqlite_table(con, 'EDGE'):
            df = pd.read_sql('SELECT condition_1, condition_2, bait_id, prey_id, pvalue, pvalue_adjusted, level, bait_id || "_" || prey_id AS interaction_id FROM EDGE_LEVEL;', con)
        else:
            df = None

        return df

    def read_monomer_qmeta(self, con):
        df = None

        if check_sqlite_table(con, 'NODE'):
            df = pd.read_sql('SELECT condition_1, condition_2, bait_id, pvalue, pvalue_adjusted, level FROM NODE_LEVEL;', con)
        else:
            df = None

        return df

    def read_baits(self, con):
        if not check_sqlite_table(con, 'NODE'):
            sys.exit("Error: Your experimental design is not supported. At least two conditions are necessary for differential analysis. Switch 'level' to 'interaction' for visualization.")

        if self.combined:
            table = 'NODE'
        else:
            table = 'NODE_LEVEL'

        if self.mode == 'quantitative':
            df = pd.read_sql('SELECT DISTINCT bait_id, min(pvalue) as pvalue FROM %s WHERE pvalue_adjusted < %s AND abs_log2fx > %s GROUP BY bait_id;' % (table, self.max_qvalue, self.min_abs_log2fx), con)

        df = df.sort_values(by=['pvalue']).reset_index()

        return df['bait_id'].values, df.index+1



import pandas as pd
import numpy as np
import scipy as sp
import click
import sqlite3
import pickle
import os
import sys

try:
    import matplotlib
    matplotlib.use('Agg')
    from matplotlib.backends.backend_pdf import PdfPages
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

from scipy.stats import gaussian_kde
from numpy import linspace, concatenate, around

from pyprophet.pyprophet import PyProphet
from pyprophet.report import save_report
from pyprophet.stats import pemp, qvalue, pi0est

from hyperopt import hp

class pyprophet:
    def __init__(self, outfile, apply_model, minimum_abundance_ratio, maximum_sec_shift, cb_decoys, xeval_fraction, xeval_num_iter, ss_initial_fdr, ss_iteration_fdr, ss_num_iter, xgb_autotune, parametric, pfdr, pi0_lambda, pi0_method, pi0_smooth_df, pi0_smooth_log_pi0, lfdr_truncate, lfdr_monotone, lfdr_transformation, lfdr_adj, lfdr_eps, plot_reports, threads, test, export_tables):

        self.outfile = outfile
        self.apply_model = apply_model
        self.classifier = 'XGBoost'
        self.xgb_autotune = xgb_autotune

        self.xgb_hyperparams = {'autotune': self.xgb_autotune, 'autotune_num_rounds': 10, 'num_boost_round': 100, 'early_stopping_rounds': 10, 'test_size': 0.33}

        self.xgb_params = {'eta': 1.0, 'gamma': 0, 'max_depth': 6, 'min_child_weight': 1, 'subsample': 1, 'colsample_bytree': 1, 'colsample_bylevel': 1, 'colsample_bynode': 1, 'lambda': 1, 'alpha': 0, 'scale_pos_weight': 1, 'objective': 'binary:logitraw', 'nthread': 1, 'eval_metric': 'auc'}

        self.xgb_params_space = {'eta': hp.uniform('eta', 0.5, 1.0), 'gamma': hp.uniform('gamma', 0.0, 0.5), 'max_depth': hp.quniform('max_depth', 2, 8, 1), 'min_child_weight': hp.quniform('min_child_weight', 1, 5, 1), 'subsample': hp.uniform('subsample', 0.5, 1.0), 'colsample_bytree': 1.0, 'colsample_bylevel': 1.0, 'colsample_bynode': 1.0, 'lambda': hp.uniform('lambda', 0.0, 1.0), 'alpha': hp.uniform('alpha', 0.0, 1.0), 'scale_pos_weight': 1.0, 'silent': 1, 'objective': 'binary:logitraw', 'nthread': 1, 'eval_metric': 'auc'}

        self.minimum_abundance_ratio = minimum_abundance_ratio
        self.maximum_sec_shift = maximum_sec_shift
        self.cb_decoys = cb_decoys
        self.xeval_fraction = xeval_fraction
        self.xeval_num_iter = xeval_num_iter
        self.ss_initial_fdr = ss_initial_fdr
        self.ss_iteration_fdr = ss_iteration_fdr
        self.ss_num_iter = ss_num_iter
        self.group_id = 'pyprophet_feature_id'
        self.parametric = parametric
        self.pfdr = pfdr
        self.pi0_lambda = pi0_lambda
        self.pi0_method = pi0_method
        self.pi0_smooth_df = pi0_smooth_df
        self.pi0_smooth_log_pi0 = pi0_smooth_log_pi0
        self.lfdr_truncate = lfdr_truncate
        self.lfdr_monotone = lfdr_monotone
        self.lfdr_transformation = lfdr_transformation
        self.lfdr_adj = lfdr_adj
        self.lfdr_eps = lfdr_eps
        self.threads = threads
        self.plot_reports = plot_reports
        self.test = test
        self.export_tables = export_tables
        self.has_learning = self.has_learning()

        # Load pretrained model if available
        if self.apply_model is not None:
            self.weights = self.load_model()

        # Learn classifier
        else:
            if self.has_learning:
                learning_data = self.read_data(learning=True)
                if self.cb_decoys:
                    click.echo("Info: Using decoys from same confidence bin for learning.")
                    self.weights = self.learn(learning_data[(learning_data['learning'] == 1)])
                else:
                    self.weights = self.learn(learning_data[(learning_data['learning'] == 1) | (learning_data['decoy'] == 1)])
            else:
                learning_data = self.read_data(learning=False)
                self.weights = self.learn(learning_data[learning_data['confidence_bin'] == learning_data['confidence_bin'].max()])
            # Store model
            self.store_model()

        # Apply classifier to full dataset
        runs = self.read_runs()

        # Apply separately for each run
        for run in runs.iterrows():
            click.echo("Info: Apply scores to condition %s and replicate %s." %(run[1]['condition_id'], run[1]['replicate_id']))
            data = self.read_data(learning=False, condition_id=run[1]['condition_id'], replicate_id=run[1]['replicate_id'])

            if self.has_learning:
                scored_data = data[data['learning'] == 0].groupby('confidence_bin', group_keys=False).apply(self.apply, condition_id=run[1]['condition_id'], replicate_id=run[1]['replicate_id'])
            else:
                scored_data = data.groupby('confidence_bin', group_keys=False).apply(self.apply, condition_id=run[1]['condition_id'], replicate_id=run[1]['replicate_id'])

            con = sqlite3.connect(outfile)
            scored_data.to_sql('FEATURE_SCORED', con, index=False, if_exists='append')
            con.close()

    def has_learning(self):
        con = sqlite3.connect(self.outfile)
        c = con.cursor()
        c.execute('SELECT count(*) FROM FEATURE WHERE learning==1;')
        if c.fetchone()[0] == 0:
            learning = False
        else:
            learning = True
        con.close()
        return learning

    def read_runs(self):
        con = sqlite3.connect(self.outfile)
        df = pd.read_sql('SELECT DISTINCT condition_id, replicate_id FROM FEATURE;', con)
        con.close()

        return df

    def read_data(self, learning=False, condition_id=None, replicate_id=None):
        con = sqlite3.connect(self.outfile)
        if learning and condition_id is None and replicate_id is None:
            df = pd.read_sql('SELECT *, condition_id || "_" || replicate_id || "_" || bait_id || "_" || prey_id || "_" || decoy AS pyprophet_feature_id, condition_id || "_" || bait_id || "_" || prey_id || "_" || decoy AS pyprophet_metafeature_id FROM FEATURE WHERE learning==1 OR decoy==1 ORDER BY pyprophet_metafeature_id;', con)
        elif condition_id is not None and replicate_id is not None:
            df = pd.read_sql('SELECT *, condition_id || "_" || replicate_id || "_" || bait_id || "_" || prey_id || "_" || decoy AS pyprophet_feature_id, condition_id || "_" || bait_id || "_" || prey_id || "_" || decoy AS pyprophet_metafeature_id FROM FEATURE WHERE learning==0 AND condition_id=="%s" AND replicate_id=="%s" ORDER BY pyprophet_metafeature_id;' % (condition_id, replicate_id), con)
        else:
            df = pd.read_sql('SELECT *, condition_id || "_" || replicate_id || "_" || bait_id || "_" || prey_id || "_" || decoy AS pyprophet_feature_id, condition_id || "_" || bait_id || "_" || prey_id || "_" || decoy AS pyprophet_metafeature_id FROM FEATURE ORDER BY pyprophet_metafeature_id;', con)
        con.close()

        # Filter according to boundaries
        df_filter = df.groupby(["bait_id","prey_id","decoy"], group_keys=False)[["var_xcorr_shift","var_abundance_ratio","var_total_abundance_ratio"]].mean(numeric_only=True).reset_index(level=["bait_id","prey_id","decoy"])

        df_filter = df_filter[(df_filter['var_xcorr_shift'] <= self.maximum_sec_shift) & (df_filter['var_abundance_ratio'] >= self.minimum_abundance_ratio) & (df_filter['var_total_abundance_ratio'] >= self.minimum_abundance_ratio)]

        df = pd.merge(df, df_filter[["bait_id","prey_id","decoy"]], on=["bait_id","prey_id","decoy"])

        # We need to generate a kickstart score for semi-supervised learning that selects for the very best interaction heterodimers: perfect shape, co-elution and overlap
        df['main_var_kickstart'] = (df['var_xcorr_shape'] * df['var_total_abundance_ratio']) / (df['var_xcorr_shift'] + 1)

        # df = df.rename(index=str, columns={"var_xcorr_shape": "main_var_xcorr_shape", "var_sec_overlap": "main_var_sec_overlap"})

        return df

    def learn(self, learning_data):
        (result, scorer, weights) = PyProphet(
            self.classifier, 
            self.xgb_hyperparams, 
            self.xgb_params, 
            self.xgb_params_space, 
            self.xeval_fraction, 
            self.xeval_num_iter, 
            self.ss_initial_fdr, 
            self.ss_iteration_fdr, 
            self.ss_num_iter, 
            self.group_id, 
            self.parametric, 
            self.pfdr, 
            self.pi0_lambda, 
            self.pi0_method, 
            self.pi0_smooth_df, 
            self.pi0_smooth_log_pi0, 
            self.lfdr_truncate, 
            self.lfdr_monotone, 
            self.lfdr_transformation, 
            self.lfdr_adj, 
            self.lfdr_eps, 
            False, 
            self.threads, 
            self.test, 
            ss_score_filter = '', 
            color_palette='normal'
        ).learn_and_apply(learning_data)

        if self.export_tables:
            file_name = os.path.splitext(os.path.basename(self.outfile))[0]+"_learn_int_scored.csv"
            result.scored_tables.to_csv(file_name, index=False)

        self.plot(result, scorer.pi0, "learning")
        self.plot_scores(result.scored_tables, "learning")

        return weights

    def store_model(self):
        con = sqlite3.connect(self.outfile)
        c = con.cursor()
        c.execute('SELECT count(name) FROM sqlite_master WHERE type="table" AND name="PYPROPHET_XGB";')
        if c.fetchone()[0] == 1:
            c.execute('DELETE FROM PYPROPHET_XGB')
        else:
            c.execute('CREATE TABLE PYPROPHET_XGB (xgb BLOB)')

        c.execute('INSERT INTO PYPROPHET_XGB VALUES(?)', [pickle.dumps(self.weights)])
        con.commit()
        con.close()

    def load_model(self):
        try:
            con = sqlite3.connect(self.apply_model)

            data = con.execute("SELECT xgb FROM PYPROPHET_XGB").fetchone()
            con.close()
        except Exception:
            import traceback
            traceback.print_exc()
            raise

        return pickle.loads(data[0])

    def apply(self, detecting_data, condition_id, replicate_id):
        (result, scorer, weights) = PyProphet(self.classifier, self.xgb_hyperparams, self.xgb_params, self.xgb_params_space, self.xeval_fraction, self.xeval_num_iter, self.ss_initial_fdr, self.ss_iteration_fdr, self.ss_num_iter, self.group_id, self.parametric, self.pfdr, self.pi0_lambda, self.pi0_method, self.pi0_smooth_df, self.pi0_smooth_log_pi0, self.lfdr_truncate, self.lfdr_monotone, self.lfdr_transformation, self.lfdr_adj, self.lfdr_eps, False, self.threads, self.test, ss_score_filter = '', color_palette='normal').apply_weights(detecting_data, self.weights)

        df = result.scored_tables[['condition_id','replicate_id','bait_id','prey_id','decoy','confidence_bin','d_score','p_value','q_value','pep']]
        df.columns = ['condition_id','replicate_id','bait_id','prey_id','decoy','confidence_bin','score','pvalue','qvalue','pep']

        if self.plot_reports:
            self.plot(result, scorer.pi0, condition_id + "_" + replicate_id + "_" + "detecting_" + str(detecting_data['confidence_bin'].values[0]))
            self.plot_scores(result.scored_tables, condition_id + "_" + replicate_id + "_" + "detecting_" + str(detecting_data['confidence_bin'].values[0]))

        return df

    def plot(self, result, pi0, tag):
        cutoffs = result.final_statistics["cutoff"].values
        svalues = result.final_statistics["svalue"].values
        qvalues = result.final_statistics["qvalue"].values

        pvalues = result.scored_tables.loc[(result.scored_tables.peak_group_rank == 1) & (result.scored_tables.decoy == 0)]["p_value"].values
        top_targets = result.scored_tables.loc[(result.scored_tables.peak_group_rank == 1) & (result.scored_tables.decoy == 0)]["d_score"].values
        top_decoys = result.scored_tables.loc[(result.scored_tables.peak_group_rank == 1) & (result.scored_tables.decoy == 1)]["d_score"].values

        save_report(os.path.splitext(os.path.basename(self.outfile))[0]+"_"+tag+".pdf", tag, top_decoys, top_targets, cutoffs, svalues, qvalues, pvalues, pi0)

    def plot_scores(self, df, tag):
        if plt is None:
            raise ImportError("Error: The matplotlib package is required to create a report.")

        out = os.path.splitext(os.path.basename(self.outfile))[0]+"_"+tag+"_scores.pdf"

        score_columns = ["d_score"] + [c for c in df.columns if c.startswith("main_var_")] + [c for c in df.columns if c.startswith("var_")]

        with PdfPages(out) as pdf:
            plt.figure(figsize=(10, 10))
            plt.subplots_adjust(hspace=.5)
            for idx in score_columns:
                top_targets = df[df["decoy"] == 0][idx]
                top_decoys = df[df["decoy"] == 1][idx]

                if not (top_targets.isnull().values.any() or top_decoys.isnull().values.any()):
                    tdensity = gaussian_kde(top_targets)
                    tdensity.covariance_factor = lambda: .25
                    tdensity._compute_covariance()
                    ddensity = gaussian_kde(top_decoys)
                    ddensity.covariance_factor = lambda: .25
                    ddensity._compute_covariance()
                    xs = linspace(min(concatenate((top_targets, top_decoys))), max(
                        concatenate((top_targets, top_decoys))), 200)

                    plt.subplot(211)
                    plt.title(idx)
                    plt.xlabel(idx)
                    plt.ylabel("# of groups")
                    plt.hist(
                        [top_targets, top_decoys], 20, color=['g', 'r'], label=['target', 'decoy'], histtype='bar')
                    plt.legend(loc=2)

                    plt.subplot(212)
                    plt.xlabel(idx)
                    plt.ylabel("density")
                    plt.plot(xs, tdensity(xs), color='g', label='target')
                    plt.plot(xs, ddensity(xs), color='r', label='decoy')
                    plt.legend(loc=2)

                    pdf.savefig()
                    plt.clf()
                    plt.close()

class combine:
    def __init__(self, outfile, pi0_lambda, pi0_method, pi0_smooth_df, pi0_smooth_log_pi0, pfdr):

        self.outfile = outfile
        self.pi0_lambda = pi0_lambda
        self.pi0_method = pi0_method
        self.pi0_smooth_df = pi0_smooth_df
        self.pi0_smooth_log_pi0 = pi0_smooth_log_pi0
        self.pfdr = pfdr

        scores = self.read()
        self.df = scores.groupby('confidence_bin', group_keys=False).apply(self.combine_scores)

    def read(self):
        con = sqlite3.connect(self.outfile)
        df = pd.read_sql('SELECT condition_id, bait_id , prey_id , decoy , confidence_bin, score, qvalue FROM FEATURE_SCORED;', con)
        con.close()

        return df

    def combine_scores(self, scores):
        combined_scores = scores.groupby(['condition_id','bait_id','prey_id','decoy','confidence_bin'], group_keys=False)['score'].mean(numeric_only=True).reset_index()

        combined_scores.loc[combined_scores['decoy'] == 0,'pvalue'] = pemp(combined_scores[combined_scores['decoy'] == 0]['score'], combined_scores[combined_scores['decoy'] == 1]['score'])

        pi0_combined = pi0est(combined_scores[combined_scores['decoy'] == 0]['pvalue'], self.pi0_lambda, self.pi0_method, self.pi0_smooth_df, self.pi0_smooth_log_pi0)['pi0']
        combined_scores.loc[combined_scores['decoy'] == 0,'qvalue'] = qvalue(combined_scores[combined_scores['decoy'] == 0]['pvalue'], pi0_combined, self.pfdr)

        click.echo("Info: Unique interactions detected before integration:")
        click.echo("%s (at q-value < 0.01)" % (scores[(scores['decoy'] == 0) & (scores['qvalue'] < 0.01)][['bait_id','prey_id']].drop_duplicates().shape[0]))
        click.echo("%s (at q-value < 0.05)" % (scores[(scores['decoy'] == 0) & (scores['qvalue'] < 0.05)][['bait_id','prey_id']].drop_duplicates().shape[0]))
        click.echo("%s (at q-value < 0.1)" % (scores[(scores['decoy'] == 0) & (scores['qvalue'] < 0.1)][['bait_id','prey_id']].drop_duplicates().shape[0]))
        click.echo("%s (at q-value < 0.2)" % (scores[(scores['decoy'] == 0) & (scores['qvalue'] < 0.2)][['bait_id','prey_id']].drop_duplicates().shape[0]))
        click.echo("%s (at q-value < 0.5)" % (scores[(scores['decoy'] == 0) & (scores['qvalue'] < 0.5)][['bait_id','prey_id']].drop_duplicates().shape[0]))

        click.echo("Info: Unique interactions detected after integration:")
        click.echo("%s (at q-value < 0.01)" % (combined_scores[(combined_scores['decoy'] == 0) & (combined_scores['qvalue'] < 0.01)][['bait_id','prey_id']].drop_duplicates().shape[0]))
        click.echo("%s (at q-value < 0.05)" % (combined_scores[(combined_scores['decoy'] == 0) & (combined_scores['qvalue'] < 0.05)][['bait_id','prey_id']].drop_duplicates().shape[0]))
        click.echo("%s (at q-value < 0.1)" % (combined_scores[(combined_scores['decoy'] == 0) & (combined_scores['qvalue'] < 0.1)][['bait_id','prey_id']].drop_duplicates().shape[0]))
        click.echo("%s (at q-value < 0.2)" % (combined_scores[(combined_scores['decoy'] == 0) & (combined_scores['qvalue'] < 0.2)][['bait_id','prey_id']].drop_duplicates().shape[0]))
        click.echo("%s (at q-value < 0.5)" % (combined_scores[(combined_scores['decoy'] == 0) & (combined_scores['qvalue'] < 0.5)][['bait_id','prey_id']].drop_duplicates().shape[0]))
        click.echo("Info: Combined pi0: %s." % pi0_combined)

        return combined_scores
